# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests the S3 backend store"""

from glance_store._drivers import s3
from glance_store.tests import base
from glance_store.tests.unit import test_s3_store_base
from glance_store.tests.unit import test_store_capabilities

S3_CONF = {
    's3_store_access_key': 'user',
    's3_store_secret_key': 'key',
    's3_store_region_name': '',
    's3_store_host': 'localhost',
    's3_store_bucket': 'glance',
    's3_store_large_object_size': 9,  # over 9MB is large
    's3_store_large_object_chunk_size': 6,  # part size is 6MB
    's3_store_enable_data_integrity_protection': False,
    's3_store_request_checksum_calculation': 'when_required',
    's3_store_response_checksum_validation': 'when_required',
}


class TestStore(base.StoreBaseTest,
                test_s3_store_base.TestS3StoreBase,
                test_store_capabilities.TestStoreCapabilitiesChecking):

    def setUp(self):
        """Establish a clean test environment."""
        super(TestStore, self).setUp()
        # Set default values for multistore and backend
        self.multistore = False
        self.backend = 'glance_store'

        self.store = s3.Store(self.conf)
        self.config(**S3_CONF)
        self.store.configure()
        self.register_store_schemes(self.store, 's3')

        self.hash_algo = 'sha256'

    def test_get_invalid_bucket_name(self):
        self._test_get_invalid_bucket_name()

    def test_client_custom_region_name(self):
        self._test_client_custom_region_name()

    def test_client_custom_ca_cert_bundle(self):
        self._test_client_custom_ca_cert_bundle()

    def test_get(self):
        self._test_get()

    def test_partial_get(self):
        self._test_partial_get()

    def test_get_non_existing(self):
        self._test_get_non_existing()

    def test_add_singlepart(self):
        self._test_add_singlepart()

    def test_add_singlepart_size_exceeding_max_size(self):
        self._test_add_singlepart_size_exceeding_max_size()

    def test_add_singlepart_write_less_than_declared(self):
        self._test_add_singlepart_write_less_than_declared()

    def test_add_singlepart_bigger_than_write_chunk(self):
        self._test_add_singlepart_bigger_than_write_chunk()

    def test_add_with_verifier(self):
        self._test_add_with_verifier()

    def test_add_multipart(self):
        self._test_add_multipart()

    def test_add_multipart_size_exceeding_max_size(self):
        self._test_add_multipart_size_exceeding_max_size()

    def test_add_multipart_write_less_than_declared(self):
        self._test_add_multipart_write_less_than_declared()

    def test_add_already_existing(self):
        self._test_add_already_existing()

    def test_no_access_key(self):
        self._test_no_access_key()

    def test_no_secret_key(self):
        self._test_no_secret_key()

    def test_no_host(self):
        self._test_no_host()

    def test_no_bucket(self):
        self._test_no_bucket()

    def test_delete_non_existing(self):
        self._test_delete_non_existing()

    def test_get_s3_good_location(self):
        self._test_get_s3_good_location()

    def test_get_my_object_storage_location(self):
        self._test_get_my_object_storage_location()

    def test_config_with_data_integrity_protection_disabled(self):
        self._test_config_with_data_integrity_protection_disabled()

    def test_config_with_data_integrity_protection_enabled(self):
        self._test_config_with_data_integrity_protection_enabled()

    def test_config_fallback_for_old_boto3(self):
        self._test_config_fallback_for_old_boto3()



# Copyright 2025 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import hashlib
import io
from unittest import mock
import uuid

import boto3
import botocore
from botocore import exceptions as boto_exceptions
from botocore import stub
from oslo_utils import units

from glance_store._drivers import s3
from glance_store import capabilities
from glance_store import exceptions
from glance_store import location

# Constants
FAKE_UUID = str(uuid.uuid4())
FIVE_KB = 5 * units.Ki
S3_CONF = {
    's3_store_access_key': 'user',
    's3_store_secret_key': 'key',
    's3_store_region_name': '',
    's3_store_host': 'localhost',
    's3_store_bucket': 'glance',
    's3_store_large_object_size': 9,  # over 9MB is large
    's3_store_large_object_chunk_size': 6,  # part size is 6MB
}


def format_s3_location(user, key, authurl, bucket, obj):
    """Helper method that returns a S3 store URI given the component pieces."""
    scheme = 's3'
    if authurl.startswith('https://'):
        scheme = 's3+https'
        authurl = authurl[len('https://'):]
    elif authurl.startswith('http://'):
        authurl = authurl[len('http://'):]
    authurl = authurl.strip('/')
    return "%s://%s:%s@%s/%s/%s" % (scheme, user, key, authurl, bucket, obj)


class TestS3StoreBase(object):

    def _test_get(self):
        """Test a "normal" retrieval of an image in chunks."""
        bucket, key = 'glance', str(uuid.uuid4())
        fixture_object = {
            'Body': io.BytesIO(b"*" * 5 * units.Ki),
            'ContentLength': 5 * units.Ki
        }
        fake_s3_client = botocore.session.get_session().create_client('s3')

        with stub.Stubber(fake_s3_client) as stubber:
            stubber.add_response(method='head_object',
                                 service_response={},
                                 expected_params={
                                     'Bucket': bucket,
                                     'Key': key
                                 })
            stubber.add_response(method='get_object',
                                 service_response=fixture_object,
                                 expected_params={
                                     'Bucket': bucket,
                                     'Key': key
                                 })

            with mock.patch.object(boto3.session.Session,
                                   "client") as mock_client:
                mock_client.return_value = fake_s3_client

                if self.multistore:
                    loc = location.get_location_from_uri_and_backend(
                        "s3://user:key@auth_address/%s/%s" % (bucket, key),
                        self.backend, conf=self.conf)
                else:
                    loc = location.get_location_from_uri(
                        "s3://user:key@auth_address/%s/%s" % (bucket, key),
                        conf=self.conf)
                (image_s3, image_size) = self.store.get(loc)

                self.assertEqual(5 * units.Ki, image_size)

                expected_data = b"*" * 5 * units.Ki
                data = b""

                for chunk in image_s3:
                    data += chunk
                self.assertEqual(expected_data, data)

    def _test_get_non_existing(self):
        """Test trying to retrieve a s3 that doesn't exist raises an error."""
        bucket, key = 'glance', 'no_exist'
        fake_s3_client = botocore.session.get_session().create_client('s3')

        with stub.Stubber(fake_s3_client) as stubber:
            stubber.add_client_error(method='head_object',
                                     service_error_code='404',
                                     service_message='''
                                     The specified key does not exist.
                                     ''',
                                     expected_params={
                                         'Bucket': bucket,
                                         'Key': key
                                     })

            with mock.patch.object(
                    boto3.session.Session, "client") as mock_client:
                mock_client.return_value = fake_s3_client

                uri = "s3://user:key@auth_address/%s/%s" % (bucket, key)
                if self.multistore:
                    loc = location.get_location_from_uri_and_backend(
                        uri, self.backend, conf=self.conf)
                else:
                    loc = location.get_location_from_uri(uri, conf=self.conf)
                self.assertRaises(exceptions.NotFound, self.store.get, loc)

    def _test_add_singlepart(self):
        """Test that we can add an image via the s3 backend."""
        expected_image_id = str(uuid.uuid4())
        # 5KiB is smaller than WRITE_CHUNKSIZE
        expected_s3_size = 5 * units.Ki
        expected_s3_contents = b"*" * expected_s3_size
        expected_checksum = hashlib.md5(expected_s3_contents,
                                        usedforsecurity=False).hexdigest()
        expected_multihash = hashlib.sha256(expected_s3_contents).hexdigest()
        backend_conf = getattr(self.conf, self.backend)

        expected_location = format_s3_location(
            backend_conf.s3_store_access_key,
            backend_conf.s3_store_secret_key,
            backend_conf.s3_store_host,
            backend_conf.s3_store_bucket,
            expected_image_id)
        image_s3 = io.BytesIO(expected_s3_contents)

        fake_s3_client = botocore.session.get_session().create_client('s3')

        with stub.Stubber(fake_s3_client) as stubber:
            stubber.add_response(method='head_bucket',
                                 service_response={},
                                 expected_params={
                                     'Bucket': backend_conf.s3_store_bucket
                                 })
            stubber.add_client_error(
                method='head_object', service_error_code='404',
                service_message='',
                expected_params={
                    'Bucket': backend_conf.s3_store_bucket,
                    'Key': expected_image_id})
            stubber.add_response(method='put_object',
                                 service_response={},
                                 expected_params={
                                     'Bucket': backend_conf.s3_store_bucket,
                                     'Key': expected_image_id,
                                     'Body': botocore.stub.ANY
                                 })

            with mock.patch.object(
                    boto3.session.Session, "client") as mock_client:
                mock_client.return_value = fake_s3_client

                if self.multistore:
                    loc, size, checksum, multihash, metadata = \
                        self.store.add(expected_image_id, image_s3,
                                       expected_s3_size, self.hash_algo)
                    self.assertEqual(self.backend, metadata["store"])
                else:
                    loc, size, checksum, multihash, _ = \
                        self.store.add(expected_image_id, image_s3,
                                       expected_s3_size, self.hash_algo)

                self.assertEqual(expected_location, loc)
                self.assertEqual(expected_s3_size, size)
                self.assertEqual(expected_checksum, checksum)
                self.assertEqual(expected_multihash, multihash)

    def _test_add_singlepart_size_exceeding_max_size(self):
        """Test size validation during add for singlepart upload."""
        self.store.WRITE_CHUNKSIZE = 1024
        expected_image_id = str(uuid.uuid4())
        # Provide a smaller expected size than actual data to trigger
        # size mismatch
        expected_s3_size = 5 * units.Ki
        # 5KB more than expected
        actual_s3_contents = b"*" * (expected_s3_size * 2)
        image_s3 = io.BytesIO(actual_s3_contents)
        backend_conf = getattr(self.conf, self.backend)

        fake_s3_client = botocore.session.get_session().create_client('s3')

        with stub.Stubber(fake_s3_client) as stubber:
            stubber.add_response(method='head_bucket',
                                 service_response={},
                                 expected_params={
                                     'Bucket': backend_conf.s3_store_bucket
                                 })
            stubber.add_client_error(
                method='head_object', service_error_code='404',
                service_message='', expected_params={
                    'Bucket': backend_conf.s3_store_bucket,
                    'Key': expected_image_id})
            stubber.add_response(method='put_object',
                                 service_response={},
                                 expected_params={
                                     'Bucket': backend_conf.s3_store_bucket,
                                     'Key': expected_image_id,
                                     'Body': botocore.stub.ANY
                                 })

            with mock.patch.object(
                    boto3.session.Session, "client") as mock_client:
                mock_client.return_value = fake_s3_client

                # Expect an exception due to size mismatch
                self.assertRaisesRegex(exceptions.Invalid,
                                       'Size exceeds: expected',
                                       self.store.add,
                                       expected_image_id, image_s3,
                                       expected_s3_size, self.hash_algo)

                # Verify that the stream's position reflects the number of
                # bytes read, which should be exactly at expected_file_size
                # plus the last buffer size read.
                actual_read_size = image_s3.tell()
                expected_read = expected_s3_size + self.store.WRITE_CHUNKSIZE
                self.assertEqual(actual_read_size, expected_read,
                                 "The stream was not read only up to the"
                                 " expected size.")

    def _test_add_singlepart_write_less_than_declared(self):
        """Test size validation during add for singlepart upload."""
        expected_image_id = str(uuid.uuid4())
        # Provide a smaller expected size than actual data to trigger
        # size mismatch
        expected_s3_size = 5 * units.Ki
        # 100 bytes smaller than expected
        actual_s3_contents = b"*" * (5 * units.Ki - 100)
        image_s3 = io.BytesIO(actual_s3_contents)
        backend_conf = getattr(self.conf, self.backend)

        fake_s3_client = botocore.session.get_session().create_client('s3')

        with stub.Stubber(fake_s3_client) as stubber:
            stubber.add_response(method='head_bucket',
                                 service_response={},
                                 expected_params={
                                     'Bucket': backend_conf.s3_store_bucket
                                 })
            stubber.add_client_error(
                method='head_object', service_error_code='404',
                service_message='', expected_params={
                    'Bucket': backend_conf.s3_store_bucket,
                    'Key': expected_image_id})
            stubber.add_response(method='put_object',
                                 service_response={},
                                 expected_params={
                                     'Bucket': backend_conf.s3_store_bucket,
                                     'Key': expected_image_id,
                                     'Body': botocore.stub.ANY
                                 })

            with mock.patch.object(
                    boto3.session.Session, "client") as mock_client:
                mock_client.return_value = fake_s3_client

                # Expect an exception due to size mismatch
                self.assertRaisesRegex(exceptions.Invalid,
                                       'Size mismatch: expected',
                                       self.store.add,
                                       expected_image_id, image_s3,
                                       expected_s3_size, self.hash_algo)
                # The input buffer should be fully read
                # depending on implementation
                total_input_size = len(actual_s3_contents)
                self.assertEqual(image_s3.tell(), total_input_size)

    def _test_add_singlepart_bigger_than_write_chunk(self):
        """Test that we can add a large image via the s3 backend."""
        expected_image_id = str(uuid.uuid4())
        # 8 MiB is bigger than WRITE_CHUNKSIZE(=5MiB),
        # but smaller than s3_store_large_object_size
        expected_s3_size = 8 * units.Mi
        expected_s3_contents = b"*" * expected_s3_size
        expected_checksum = hashlib.md5(expected_s3_contents,
                                        usedforsecurity=False).hexdigest()
        expected_multihash = hashlib.sha256(expected_s3_contents).hexdigest()
        backend_conf = getattr(self.conf, self.backend)

        expected_location = format_s3_location(
            backend_conf.s3_store_access_key,
            backend_conf.s3_store_secret_key,
            backend_conf.s3_store_host,
            backend_conf.s3_store_bucket,
            expected_image_id)
        image_s3 = io.BytesIO(expected_s3_contents)

        fake_s3_client = botocore.session.get_session().create_client('s3')

        with stub.Stubber(fake_s3_client) as stubber:
            stubber.add_response(method='head_bucket',
                                 service_response={},
                                 expected_params={
                                     'Bucket': backend_conf.s3_store_bucket
                                 })
            stubber.add_client_error(
                method='head_object', service_error_code='404',
                service_message='', expected_params={
                    'Bucket': backend_conf.s3_store_bucket,
                    'Key': expected_image_id})
            stubber.add_response(method='put_object',
                                 service_response={},
                                 expected_params={
                                     'Bucket': backend_conf.s3_store_bucket,
                                     'Key': expected_image_id,
                                     'Body': botocore.stub.ANY
                                 })

            with mock.patch.object(
                    boto3.session.Session, "client") as mock_client:
                mock_client.return_value = fake_s3_client

                if self.multistore:
                    loc, size, checksum, multihash, metadata = \
                        self.store.add(
                            expected_image_id, image_s3, expected_s3_size,
                            self.hash_algo)
                    self.assertEqual(self.backend, metadata["store"])
                else:
                    loc, size, checksum, multihash, _ = \
                        self.store.add(expected_image_id, image_s3,
                                       expected_s3_size, self.hash_algo)

                self.assertEqual(expected_location, loc)
                self.assertEqual(expected_s3_size, size)
                self.assertEqual(expected_checksum, checksum)
                self.assertEqual(expected_multihash, multihash)

    def _test_add_with_verifier(self):
        """Assert 'verifier.update' is called when verifier is provided."""
        expected_image_id = str(uuid.uuid4())
        expected_s3_size = 5 * units.Ki
        expected_s3_contents = b"*" * expected_s3_size
        image_s3 = io.BytesIO(expected_s3_contents)

        fake_s3_client = botocore.session.get_session().create_client('s3')
        verifier = mock.MagicMock(name='mock_verifier')

        with stub.Stubber(fake_s3_client) as stubber:
            stubber.add_response(method='head_bucket', service_response={})
            stubber.add_client_error(method='head_object',
                                     service_error_code='404',
                                     service_message='')
            stubber.add_response(method='put_object', service_response={})

            with mock.patch.object(
                    boto3.session.Session, "client") as mock_client:
                mock_client.return_value = fake_s3_client
                self.store.add(expected_image_id, image_s3,
                               expected_s3_size, self.hash_algo,
                               verifier=verifier)
        verifier.update.assert_called_with(expected_s3_contents)

    def _test_add_multipart(self):
        """Test that we can add an image via the s3 backend."""
        expected_image_id = str(uuid.uuid4())
        expected_s3_size = 16 * units.Mi
        expected_s3_contents = b"*" * expected_s3_size
        expected_checksum = hashlib.md5(expected_s3_contents,
                                        usedforsecurity=False).hexdigest()
        expected_multihash = hashlib.sha256(expected_s3_contents).hexdigest()
        backend_conf = getattr(self.conf, self.backend)

        expected_location = format_s3_location(
            backend_conf.s3_store_access_key,
            backend_conf.s3_store_secret_key,
            backend_conf.s3_store_host,
            backend_conf.s3_store_bucket,
            expected_image_id)
        image_s3 = io.BytesIO(expected_s3_contents)

        fake_s3_client = botocore.session.get_session().create_client('s3')

        num_parts = 3  # image size is 16MB and chunk size is 6MB
        with stub.Stubber(fake_s3_client) as stubber:
            stubber.add_response(method='head_bucket',
                                 service_response={},
                                 expected_params={
                                     'Bucket': backend_conf.s3_store_bucket
                                 })
            stubber.add_client_error(
                method='head_object', service_error_code='404',
                service_message='', expected_params={
                    'Bucket': backend_conf.s3_store_bucket,
                    'Key': expected_image_id})
            stubber.add_response(method='create_multipart_upload',
                                 service_response={
                                     "Bucket": backend_conf.s3_store_bucket,
                                     "Key": expected_image_id,
                                     "UploadId": 'UploadId'
                                 },
                                 expected_params={
                                     "Bucket": backend_conf.s3_store_bucket,
                                     "Key": expected_image_id,
                                 })
            parts = []
            remaining_image_size = expected_s3_size
            large_chunk_size = backend_conf.s3_store_large_object_chunk_size
            chunk_size = large_chunk_size * units.Mi
            for i in range(num_parts):
                part_number = i + 1
                stubber.add_response(
                    method='upload_part',
                    service_response={'ETag': 'ETag'},
                    expected_params={
                        "Bucket": backend_conf.s3_store_bucket,
                        "Key": expected_image_id,
                        "Body": botocore.stub.ANY,
                        'ContentLength': chunk_size,
                        "PartNumber": part_number,
                        "UploadId": 'UploadId'
                    })
                parts.append({'ETag': 'ETag', 'PartNumber': part_number})

                remaining_image_size -= chunk_size
                if remaining_image_size < chunk_size:
                    chunk_size = remaining_image_size

            stubber.add_response(method='complete_multipart_upload',
                                 service_response={
                                     "Bucket": backend_conf.s3_store_bucket,
                                     "Key": expected_image_id,
                                     'ETag': 'ETag'
                                 },
                                 expected_params={
                                     "Bucket": backend_conf.s3_store_bucket,
                                     "Key": expected_image_id,
                                     "MultipartUpload": {
                                         "Parts": parts
                                     },
                                     "UploadId": 'UploadId'
                                 })

            with mock.patch.object(
                    boto3.session.Session, "client") as mock_client:
                mock_client.return_value = fake_s3_client

                if self.multistore:
                    loc, size, checksum, multihash, metadata = \
                        self.store.add(expected_image_id, image_s3,
                                       expected_s3_size,
                                       self.hash_algo)
                    self.assertEqual(self.backend, metadata["store"])
                else:
                    loc, size, checksum, multihash, _ = \
                        self.store.add(expected_image_id, image_s3,
                                       expected_s3_size,
                                       self.hash_algo)

                self.assertEqual(expected_location, loc)
                self.assertEqual(expected_s3_size, size)
                self.assertEqual(expected_checksum, checksum)
                self.assertEqual(expected_multihash, multihash)

    def _test_add_multipart_size_exceeding_max_size(self):
        """Test size validation during multipart upload."""
        expected_image_id = str(uuid.uuid4())
        expected_s3_size = 15 * units.Mi
        self.store.WRITE_CHUNKSIZE = 5 * units.Mi
        # Actual data is larger than expected to trigger size validation
        # failure, 5MB larger
        total_size = expected_s3_size + self.store.WRITE_CHUNKSIZE
        expected_s3_contents = b"*" * total_size
        image_s3 = io.BytesIO(expected_s3_contents)
        conf = getattr(self.conf, self.backend)

        fake_s3_client = botocore.session.get_session().create_client('s3')
        # Patch abort_multipart_upload on the client to a Mock
        fake_s3_client.abort_multipart_upload = mock.Mock()
        with stub.Stubber(fake_s3_client) as stubber:
            stubber.add_response(method='head_bucket',
                                 service_response={},
                                 expected_params={
                                     'Bucket': conf.s3_store_bucket
                                 })
            stubber.add_client_error(method='head_object',
                                     service_error_code='404',
                                     service_message='',
                                     expected_params={
                                         'Bucket': conf.s3_store_bucket,
                                         'Key': expected_image_id
                                     })
            stubber.add_response(method='create_multipart_upload',
                                 service_response={
                                     "Bucket": conf.s3_store_bucket,
                                     "Key": expected_image_id,
                                     "UploadId": 'UploadId'
                                 },
                                 expected_params={
                                     "Bucket": conf.s3_store_bucket,
                                     "Key": expected_image_id,
                                 })
            stubber.add_response(method='abort_multipart_upload',
                                 service_response={},
                                 expected_params={
                                     'Bucket': conf.s3_store_bucket,
                                     'Key': expected_image_id,
                                     'UploadId': 'UploadId'
                                 })

            with mock.patch.object(
                    boto3.session.Session, "client") as mock_client:
                mock_client.return_value = fake_s3_client
                # Expect an exception due to size mismatch
                self.assertRaisesRegex(
                    exceptions.Invalid, "Size exceeds: expected",
                    self.store.add, expected_image_id, image_s3,
                    expected_s3_size, self.hash_algo)

                # Verify that the stream's position reflects the number of
                # bytes read, which should be exactly at expected_file_size
                # plus the last buffer size read.
                actual_read_size = image_s3.tell()
                self.assertEqual(actual_read_size, total_size,
                                 "The stream was not read only up to the "
                                 "expected size.")

                # Assert that abort_multipart_upload was called
                fake_s3_client.abort_multipart_upload.assert_called_once_with(
                    Bucket=conf.s3_store_bucket,
                    Key=expected_image_id,
                    UploadId='UploadId'
                )

    def _test_add_multipart_write_less_than_declared(self):
        """Test size validation during multipart upload."""
        expected_image_id = str(uuid.uuid4())
        expected_s3_size = 16 * units.Mi
        # Actual data is less than expected to trigger size validation
        # failure, 1MB less
        expected_s3_contents = b"*" * (expected_s3_size - 1 * units.Mi)
        image_s3 = io.BytesIO(expected_s3_contents)
        conf = getattr(self.conf, self.backend)

        fake_s3_client = botocore.session.get_session().create_client('s3')
        # Patch abort_multipart_upload on the client to a Mock
        fake_s3_client.abort_multipart_upload = mock.Mock()
        with stub.Stubber(fake_s3_client) as stubber:
            stubber.add_response(method='head_bucket',
                                 service_response={},
                                 expected_params={
                                     'Bucket': conf.s3_store_bucket
                                 })
            stubber.add_client_error(method='head_object',
                                     service_error_code='404',
                                     service_message='',
                                     expected_params={
                                         'Bucket': conf.s3_store_bucket,
                                         'Key': expected_image_id
                                     })
            stubber.add_response(method='create_multipart_upload',
                                 service_response={
                                     "Bucket": conf.s3_store_bucket,
                                     "Key": expected_image_id,
                                     "UploadId": 'UploadId'
                                 },
                                 expected_params={
                                     "Bucket": conf.s3_store_bucket,
                                     "Key": expected_image_id,
                                 })
            stubber.add_response(method='abort_multipart_upload',
                                 service_response={},
                                 expected_params={
                                     'Bucket': conf.s3_store_bucket,
                                     'Key': expected_image_id,
                                     'UploadId': 'UploadId'
                                 })

            with mock.patch.object(
                    boto3.session.Session, "client") as mock_client:
                mock_client.return_value = fake_s3_client
                # Expect an exception due to size mismatch
                self.assertRaisesRegex(exceptions.Invalid,
                                       'Size mismatch: expected',
                                       self.store.add,
                                       expected_image_id, image_s3,
                                       expected_s3_size, self.hash_algo)

                # The input buffer should be fully read depending on
                # implementation
                total_input_size = len(expected_s3_contents)
                self.assertEqual(image_s3.tell(), total_input_size)

                # Assert that abort_multipart_upload was called
                fake_s3_client.abort_multipart_upload.assert_called_once_with(
                    Bucket=conf.s3_store_bucket,
                    Key=expected_image_id,
                    UploadId='UploadId'
                )

    def _test_add_already_existing(self):
        """Tests that adding an image with an existing identifier
        raises an appropriate exception
        """
        image_s3 = io.BytesIO(b"never_gonna_make_it")
        fake_s3_client = botocore.session.get_session().create_client('s3')

        with stub.Stubber(fake_s3_client) as stubber:
            stubber.add_response(method='head_bucket', service_response={})
            stubber.add_response(method='head_object', service_response={})

            with mock.patch.object(
                    boto3.session.Session, "client") as mock_client:
                mock_client.return_value = fake_s3_client
                self.assertRaises(exceptions.Duplicate, self.store.add,
                                  str(uuid.uuid4()), image_s3, 0,
                                  self.hash_algo)

    def _test_delete_non_existing(self):
        """Test trying to delete a s3 that doesn't exist raises an error."""
        bucket, key = 'glance', 'no_exist'
        fake_s3_client = botocore.session.get_session().create_client('s3')

        with stub.Stubber(fake_s3_client) as stubber:
            stubber.add_client_error(method='head_object',
                                     service_error_code='404',
                                     service_message='''
                                     The specified key does not exist.
                                     ''',
                                     expected_params={
                                         'Bucket': bucket,
                                         'Key': key
                                     })
            fake_s3_client.head_bucket = mock.MagicMock()

            with mock.patch.object(
                    boto3.session.Session, "client") as mock_client:
                mock_client.return_value = fake_s3_client

                uri = "s3://user:key@auth_address/%s/%s" % (bucket, key)
                if self.multistore:
                    loc = location.get_location_from_uri_and_backend(
                        uri, self.backend, conf=self.conf)
                else:
                    loc = location.get_location_from_uri(uri, conf=self.conf)
                self.assertRaises(exceptions.NotFound, self.store.delete, loc)

    def _do_test_get_s3_location(self, host, loc):
        """Helper method to test S3 location derivation"""
        self.assertEqual(s3.get_s3_location(host), loc)
        self.assertEqual(s3.get_s3_location(host + '/'), loc)
        self.assertEqual(s3.get_s3_location(host + ':80'), loc)
        self.assertEqual(s3.get_s3_location(host + ':80/'), loc)
        self.assertEqual(s3.get_s3_location('http://' + host), loc)
        self.assertEqual(s3.get_s3_location('http://' + host + '/'), loc)
        self.assertEqual(s3.get_s3_location('http://' + host + ':80'), loc)
        self.assertEqual(s3.get_s3_location('http://' + host + ':80/'), loc)
        self.assertEqual(s3.get_s3_location('https://' + host), loc)
        self.assertEqual(s3.get_s3_location('https://' + host + '/'), loc)
        self.assertEqual(s3.get_s3_location('https://' + host + ':80'), loc)
        self.assertEqual(s3.get_s3_location('https://' + host + ':80/'), loc)

    def _test_get_invalid_bucket_name(self):
        """Test that invalid bucket names raise appropriate exceptions"""
        if self.multistore:
            self.config(s3_store_bucket_url_format='virtual',
                        group=self.backend)
            invalid_buckets = ['not.dns.compliant', 'aa', 'bucket-']
            for bucket in invalid_buckets:
                loc = location.get_location_from_uri_and_backend(
                    "s3+https://user:key@auth_address/%s/key" % bucket,
                    self.backend, conf=self.conf)
                self.assertRaises(boto_exceptions.InvalidDNSNameError,
                                  self.store.get, loc)
        else:
            self.config(s3_store_bucket_url_format='virtual')
            invalid_buckets = ['not.dns.compliant', 'aa', 'bucket-']
            for bucket in invalid_buckets:
                loc = location.get_location_from_uri(
                    "s3://user:key@auth_address/%s/key" % bucket,
                    conf=self.conf)
                self.assertRaises(boto_exceptions.InvalidDNSNameError,
                                  self.store.get, loc)

    def _test_client_custom_region_name(self):
        """Test a custom s3_store_region_name in config"""
        if self.multistore:
            mock_loc = mock.MagicMock()
            mock_loc.accesskey = 'abcd'
            mock_loc.secretkey = 'efgh'
            mock_loc.bucket = 'bucket1'

            with mock.patch.object(
                    boto3.session.Session, "client") as mock_client:
                self.store._create_s3_client(mock_loc)
                mock_client.assert_called_with(
                    config=mock.ANY,
                    endpoint_url='https://s3-region1.com',
                    region_name='custom_region_name',
                    service_name='s3',
                    use_ssl=False,
                    verify='path/to/cert/bundle.pem',
                )
        else:
            self.config(s3_store_host='http://example.com')
            self.config(s3_store_region_name='regionOne')
            self.config(s3_store_bucket_url_format='path')
            self.store.configure()

            mock_loc = mock.MagicMock()
            mock_loc.accesskey = 'abcd'
            mock_loc.secretkey = 'efgh'
            mock_loc.bucket = 'bucket1'

            with mock.patch.object(
                    boto3.session.Session, "client") as mock_client:
                self.store._create_s3_client(mock_loc)
                mock_client.assert_called_with(
                    config=mock.ANY,
                    endpoint_url='http://example.com',
                    region_name='regionOne',
                    service_name='s3',
                    use_ssl=False,
                    verify=None,
                )

    def _test_client_custom_ca_cert_bundle(self):
        """Test a custom s3_store_cacert in config"""
        if self.multistore:
            mock_loc = mock.MagicMock()
            mock_loc.accesskey = 'abcd'
            mock_loc.secretkey = 'efgh'
            mock_loc.bucket = 'bucket1'

            with mock.patch.object(
                    boto3.session.Session, "client") as mock_client:
                self.store._create_s3_client(mock_loc)
                mock_client.assert_called_with(
                    config=mock.ANY,
                    endpoint_url='https://s3-region1.com',
                    region_name='custom_region_name',
                    service_name='s3',
                    use_ssl=False,
                    verify='path/to/cert/bundle.pem',
                )
        else:
            self.config(s3_store_host='http://example.com')
            self.config(s3_store_cacert='path/to/cert/bundle.pem')
            self.config(s3_store_bucket_url_format='path')
            self.store.configure()

            mock_loc = mock.MagicMock()
            mock_loc.accesskey = 'abcd'
            mock_loc.secretkey = 'efgh'
            mock_loc.bucket = 'bucket1'

            with mock.patch.object(
                    boto3.session.Session, "client") as mock_client:
                self.store._create_s3_client(mock_loc)
                mock_client.assert_called_with(
                    config=mock.ANY,
                    endpoint_url='http://example.com',
                    region_name=None,
                    service_name='s3',
                    use_ssl=False,
                    verify='path/to/cert/bundle.pem',
                )

    def _test_partial_get(self):
        """Test that partial get operations raise appropriate exceptions"""
        if self.multistore:
            loc = location.get_location_from_uri_and_backend(
                "s3+https://user:key@auth_address/glance/%s" % FAKE_UUID,
                self.backend, conf=self.conf)
        else:
            loc = location.get_location_from_uri(
                "s3://user:key@auth_address/glance/%s" % FAKE_UUID,
                conf=self.conf)
        self.assertRaises(exceptions.StoreRandomGetNotSupported,
                          self.store.get, loc, chunk_size=1)

    def _test_get_s3_good_location(self):
        """Test that the s3 location can be derived from the host"""
        good_locations = [
            ('s3.amazonaws.com', ''),
            ('s3-us-east-1.amazonaws.com', 'us-east-1'),
            ('s3-us-east-2.amazonaws.com', 'us-east-2'),
            ('s3-us-west-1.amazonaws.com', 'us-west-1'),
            ('s3-us-west-2.amazonaws.com', 'us-west-2'),
            ('s3-ap-east-1.amazonaws.com', 'ap-east-1'),
            ('s3-ap-south-1.amazonaws.com', 'ap-south-1'),
            ('s3-ap-northeast-1.amazonaws.com', 'ap-northeast-1'),
            ('s3-ap-northeast-2.amazonaws.com', 'ap-northeast-2'),
            ('s3-ap-northeast-3.amazonaws.com', 'ap-northeast-3'),
            ('s3-ap-southeast-1.amazonaws.com', 'ap-southeast-1'),
            ('s3-ap-southeast-2.amazonaws.com', 'ap-southeast-2'),
            ('s3-ca-central-1.amazonaws.com', 'ca-central-1'),
            ('s3-cn-north-1.amazonaws.com.cn', 'cn-north-1'),
            ('s3-cn-northwest-1.amazonaws.com.cn', 'cn-northwest-1'),
            ('s3-eu-central-1.amazonaws.com', 'eu-central-1'),
            ('s3-eu-west-1.amazonaws.com', 'eu-west-1'),
            ('s3-eu-west-2.amazonaws.com', 'eu-west-2'),
            ('s3-eu-west-3.amazonaws.com', 'eu-west-3'),
            ('s3-eu-north-1.amazonaws.com', 'eu-north-1'),
            ('s3-sa-east-1.amazonaws.com', 'sa-east-1'),
        ]
        for (url, expected) in good_locations:
            self._do_test_get_s3_location(url, expected)

    def _test_get_my_object_storage_location(self):
        """Test that the my object storage location convert to ''"""
        my_object_storage_locations = [
            ('my-object-storage.com', ''),
            ('s3-my-object.jp', ''),
            ('192.168.100.12', ''),
        ]
        for (url, expected) in my_object_storage_locations:
            self._do_test_get_s3_location(url, expected)

    def _test_no_access_key(self):
        """Tests that options without access key disables the add method"""
        self.assertTrue(self._option_required(
            's3_store_access_key'))

    def _test_no_secret_key(self):
        """Tests that options without secret key disables the add method"""
        self.assertTrue(self._option_required(
            's3_store_secret_key'))

    def _test_no_host(self):
        """Tests that options without host disables the add method"""
        self.assertTrue(self._option_required(
            's3_store_host'))

    def _test_no_bucket(self):
        """Tests that options without bucket name disables the add method"""
        self.assertTrue(self._option_required(
            's3_store_bucket'))

    def _option_required(self, key):
        """Helper method to test if a configuration option is required"""
        conf = S3_CONF.copy()
        conf[key] = None

        try:
            self.config(**conf, group=self.backend)
            self.store = s3.Store(self.conf)
            self.store.configure()
            return not self.store.is_capable(
                capabilities.BitMasks.WRITE_ACCESS)
        except Exception:
            return False

    def _test_add_different_backend(self):
        """Test adding to a different backend in multistore mode"""
        if not self.multistore:
            self.skipTest("This test is only for multistore mode")

        # Switch to different backend
        self.store = s3.Store(self.conf, backend="s3_region2")
        self.store.configure()
        self.register_store_backend_schemes(self.store, 's3', 's3_region2')

        expected_image_id = str(uuid.uuid4())
        expected_s3_size = FIVE_KB
        expected_s3_contents = b"*" * expected_s3_size
        expected_checksum = hashlib.md5(expected_s3_contents,
                                        usedforsecurity=False).hexdigest()
        expected_multihash = hashlib.sha256(expected_s3_contents).hexdigest()
        expected_location = format_s3_location(
            S3_CONF['s3_store_access_key'],
            S3_CONF['s3_store_secret_key'],
            'http://s3-region2.com',
            S3_CONF['s3_store_bucket'],
            expected_image_id)
        image_s3 = io.BytesIO(expected_s3_contents)

        fake_s3_client = botocore.session.get_session().create_client('s3')

        with stub.Stubber(fake_s3_client) as stubber:
            stubber.add_response(method='head_bucket',
                                 service_response={},
                                 expected_params={
                                     'Bucket': S3_CONF['s3_store_bucket']
                                 })
            stubber.add_client_error(method='head_object',
                                     service_error_code='404',
                                     service_message='',
                                     expected_params={
                                         'Bucket': S3_CONF['s3_store_bucket'],
                                         'Key': expected_image_id
                                     })
            stubber.add_response(method='put_object',
                                 service_response={},
                                 expected_params={
                                     'Bucket': S3_CONF['s3_store_bucket'],
                                     'Key': expected_image_id,
                                     'Body': botocore.stub.ANY
                                 })

            with mock.patch.object(
                    boto3.session.Session, "client") as mock_client:
                mock_client.return_value = fake_s3_client
                loc, size, checksum, multihash, metadata = \
                    self.store.add(expected_image_id, image_s3,
                                   expected_s3_size,
                                   self.hash_algo)
                self.assertEqual("s3_region2", metadata["store"])

                self.assertEqual(expected_location, loc)
                self.assertEqual(expected_s3_size, size)
                self.assertEqual(expected_checksum, checksum)
                self.assertEqual(expected_multihash, multihash)

    def _test_config_with_data_integrity_protection_disabled(self):
        """Test Config creation when data integrity protection is disabled."""
        # Explicitly disable data integrity protection
        self.config(s3_store_enable_data_integrity_protection=False,
                    group=self.backend)
        self.store.configure()

        # Create a mock location with required attributes
        mock_loc = mock.Mock()
        mock_loc.accesskey = 'access_key'
        mock_loc.secretkey = 'secret_key'
        mock_loc.bucket = 'test_bucket'

        with mock.patch('botocore.client.Config') as mock_config, \
                mock.patch('boto3.session.Session.client'):
            mock_config.return_value = mock.Mock()
            self.store._create_s3_client(mock_loc)

            # When data integrity protection is disabled,
            # both checksum options should be 'when_required'
            mock_config.assert_called_once()
            call_args = mock_config.call_args
            self.assertEqual('when_required',
                             call_args.kwargs['request_checksum_calculation'])
            self.assertEqual('when_required',
                             call_args.kwargs['response_checksum_validation'])

    def _test_config_with_data_integrity_protection_enabled(self):
        """Test Config creation when data integrity protection is enabled."""
        # Enable data integrity protection and set custom checksum options
        self.config(s3_store_enable_data_integrity_protection=True,
                    s3_store_request_checksum_calculation='when_supported',
                    s3_store_response_checksum_validation='when_supported',
                    group=self.backend)
        self.store.configure()

        # Create a mock location with required attributes
        mock_loc = mock.Mock()
        mock_loc.accesskey = 'access_key'
        mock_loc.secretkey = 'secret_key'
        mock_loc.bucket = 'test_bucket'

        with mock.patch('botocore.client.Config') as mock_config, \
                mock.patch('boto3.session.Session.client'):
            mock_config.return_value = mock.Mock()
            self.store._create_s3_client(mock_loc)

            # When data integrity protection is enabled,
            # checksum options should use the configured values
            mock_config.assert_called_once()
            call_args = mock_config.call_args
            self.assertEqual('when_supported',
                             call_args.kwargs['request_checksum_calculation'])
            self.assertEqual('when_supported',
                             call_args.kwargs['response_checksum_validation'])

    def _test_config_fallback_for_old_boto3(self):
        """Test Config fallback when boto3 < 1.36.0 (no checksum support)."""
        # Create a mock location with required attributes
        mock_loc = mock.Mock()
        mock_loc.accesskey = 'access_key'
        mock_loc.secretkey = 'secret_key'
        mock_loc.bucket = 'test_bucket'

        # Simulate old boto3 version by making Config raise TypeError
        # when checksum parameters are passed
        def config_side_effect(*args, **kwargs):
            if 'request_checksum_calculation' in kwargs:
                raise TypeError("Unexpected keyword argument")
            return mock.Mock()

        with mock.patch('botocore.client.Config',
                        side_effect=config_side_effect) as mock_config, \
                mock.patch('boto3.session.Session.client'):
            self.store._create_s3_client(mock_loc)

            # Should have been called twice: once with checksum params
            # (which failed), then without them
            self.assertEqual(2, mock_config.call_count)

            # First call should have checksum parameters
            first_call_kwargs = mock_config.call_args_list[0].kwargs
            self.assertIn('request_checksum_calculation',
                          first_call_kwargs)
            self.assertIn('response_checksum_validation',
                          first_call_kwargs)

            # Second call should not have checksum parameters
            second_call_kwargs = mock_config.call_args_list[1].kwargs
            self.assertNotIn('request_checksum_calculation',
                             second_call_kwargs)
            self.assertNotIn('response_checksum_validation',
                             second_call_kwargs)



# Copyright 2023 RedHat Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys
from unittest import mock

import ddt

from glance_store._drivers.cinder import base
from glance_store._drivers.cinder import scaleio
from glance_store.tests import base as test_base

sys.modules['glance_store.common.fs_mount'] = mock.Mock()
from glance_store._drivers.cinder import store as cinder # noqa
from glance_store._drivers.cinder import nfs # noqa


@ddt.ddt
class TestConnectorBase(test_base.StoreBaseTest):
    @ddt.data(
        ('iscsi', base.BaseBrickConnectorInterface),
        ('nfs', nfs.NfsBrickConnector),
        ('scaleio', scaleio.ScaleIOBrickConnector),
    )
    @ddt.unpack
    def test_factory(self, protocol, expected_class):
        connector_class = base.factory(
            connection_info={'driver_volume_type': protocol})
        self.assertIsInstance(connector_class, expected_class)


class TestBaseBrickConnectorInterface(test_base.StoreBaseTest):

    def get_connection_info(self):
        """Return iSCSI connection information"""
        return {
            'target_discovered': False,
            'target_portal': '0.0.0.0:3260',
            'target_iqn': 'iqn.2010-10.org.openstack:volume-fake-vol',
            'target_lun': 0,
            'volume_id': '007dedb8-ddc0-445c-88f1-d07acbe4efcb',
            'auth_method': 'CHAP',
            'auth_username': '2ttANgVaDRqxtMNK3hUj',
            'auth_password': 'fake-password',
            'encrypted': False,
            'qos_specs': None,
            'access_mode': 'rw',
            'cacheable': False,
            'driver_volume_type': 'iscsi',
            'attachment_id': '7f45b2fe-111a-42df-be3e-f02b312ad8ea'}

    def setUp(self, connection_info={}, **kwargs):
        super().setUp()
        self.connection_info = connection_info or self.get_connection_info()
        self.root_helper = 'fake_rootwrap'
        self.use_multipath = False
        self.properties = {
            'connection_info': self.connection_info,
            'root_helper': self.root_helper,
            'use_multipath': self.use_multipath}
        self.properties.update(kwargs)
        self.mock_object(base.connector.InitiatorConnector, 'factory')
        self.connector = base.factory(**self.properties)

    def mock_object(self, obj, attr_name, *args, **kwargs):
        """Use python mock to mock an object attribute

        Mocks the specified objects attribute with the given value.
        Automatically performs 'addCleanup' for the mock.
        """
        patcher = mock.patch.object(obj, attr_name, *args, **kwargs)
        result = patcher.start()
        self.addCleanup(patcher.stop)
        return result

    def test_connect_volume(self):
        if self.connection_info['driver_volume_type'] == 'nfs':
            self.skipTest('NFS tests have custom implementation of this '
                          'method.')
        fake_vol = mock.MagicMock()
        fake_path = {'path': 'fake_dev_path'}
        self.mock_object(self.connector.conn, 'connect_volume',
                         return_value=fake_path)
        fake_dev_path = self.connector.connect_volume(fake_vol)
        self.connector.conn.connect_volume.assert_called_once_with(
            self.connector.connection_info)
        self.assertEqual(fake_path['path'], fake_dev_path['path'])

    def test_disconnect_volume(self):
        fake_device = 'fake_dev_path'
        self.mock_object(self.connector.conn, 'disconnect_volume')
        self.connector.disconnect_volume(fake_device)
        self.connector.conn.disconnect_volume.assert_called_once_with(
            self.connection_info, fake_device, force=True)

    def test_extend_volume(self):
        self.mock_object(self.connector.conn, 'extend_volume')
        self.connector.extend_volume()
        self.connector.conn.extend_volume.assert_called_once_with(
            self.connection_info)

    def test_yield_path(self):
        fake_vol = mock.MagicMock()
        fake_device = 'fake_dev_path'
        fake_dev_path = self.connector.yield_path(fake_vol, fake_device)
        self.assertEqual(fake_device, fake_dev_path)



# Copyright 2018 RedHat Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import hashlib
import logging

from oslo_config import cfg
from oslo_utils import units
from stevedore import driver
from stevedore import extension

from glance_store import capabilities
from glance_store import exceptions
from glance_store.i18n import _
from glance_store import location


CONF = cfg.CONF
LOG = logging.getLogger(__name__)

_STORE_OPTS = [
    cfg.StrOpt('default_backend',
               help=_("""
The store identifier for the default backend in which data will be
stored.

The value must be defined as one of the keys in the dict defined
by the ``enabled_backends`` configuration option in the DEFAULT
configuration group.

If a value is not defined for this option:

* the consuming service may refuse to start
* store_add calls that do not specify a specific backend will
  raise a ``glance_store.exceptions.UnknownScheme`` exception

Related Options:
    * enabled_backends

""")),
]

FS_CONF_DATADIR_HELP = """
Directory of which the reserved store {} uses.

Possible values:
    * A valid path to a directory

Refer to [glance_store]/filesystem store config opts for more details.
"""

FS_CONF_CHUNKSIZE_HELP = """
Chunk size, in bytes to be used by reserved store {}.

The chunk size used when reading or writing image files. Raising this value
may improve the throughput but it may also slightly increase the memory usage
when handling a large number of requests.

Possible Values:
    * Any positive integer value

"""


_STORE_CFG_GROUP = 'glance_store'
_RESERVED_STORES = {}


def _list_config_opts():
    # NOTE(abhishekk): This separated approach could list
    # store options before all driver ones, which easier
    # to generate sampe config file.
    driver_opts = _list_driver_opts()
    sample_opts = [(_STORE_CFG_GROUP, _STORE_OPTS)]
    for store_entry in driver_opts:
        # NOTE(abhishekk): Do not include no_conf store
        if store_entry == "no_conf":
            continue
        sample_opts.append((store_entry, driver_opts[store_entry]))

    return sample_opts


def _list_driver_opts():
    driver_opts = {}
    mgr = extension.ExtensionManager('glance_store.drivers')
    # NOTE(zhiyan): Handle available drivers entry_points provided
    # NOTE(nikhil): Return a sorted list of drivers to ensure that the sample
    # configuration files generated by oslo config generator retain the order
    # in which the config opts appear across different runs. If this order of
    # config opts is not preserved, some downstream packagers may see a long
    # diff of the changes though not relevant as only order has changed. See
    # some more details at bug 1619487.
    drivers = sorted([ext.name for ext in mgr])
    handled_drivers = []  # Used to handle backwards-compatible entries
    for store_entry in drivers:
        driver_cls = _load_multi_store(None, store_entry, False)
        if driver_cls and driver_cls not in handled_drivers:
            if getattr(driver_cls, 'OPTIONS', None) is not None:
                driver_opts[store_entry] = driver_cls.OPTIONS
            handled_drivers.append(driver_cls)

    # NOTE(zhiyan): This separated approach could list
    # store options before all driver ones, which easier
    # to read and configure by operator.
    return driver_opts


def register_store_opts(conf, reserved_stores=None):
    LOG.debug("Registering options for group %s", _STORE_CFG_GROUP)
    conf.register_opts(_STORE_OPTS, group=_STORE_CFG_GROUP)

    configured_backends = copy.deepcopy(conf.enabled_backends)
    if reserved_stores:
        conf.enabled_backends.update(reserved_stores)
        for key in reserved_stores.keys():
            fs_conf_template = [
                cfg.StrOpt('filesystem_store_datadir',
                           default='/var/lib/glance/{}'.format(key),
                           help=FS_CONF_DATADIR_HELP.format(key)),
                cfg.MultiStrOpt('filesystem_store_datadirs',
                                help="""Not used"""),
                cfg.StrOpt('filesystem_store_metadata_file',
                           help="""Not used"""),
                cfg.IntOpt('filesystem_store_file_perm',
                           default=0,
                           help="""Not used"""),
                cfg.IntOpt('filesystem_store_chunk_size',
                           default=64 * units.Ki,
                           min=1,
                           help=FS_CONF_CHUNKSIZE_HELP.format(key)),
                cfg.BoolOpt('filesystem_thin_provisioning',
                            default=False,
                            help="""Not used""")]
            LOG.debug("Registering options for reserved store: {}".format(key))
            conf.register_opts(fs_conf_template, group=key)

    driver_opts = _list_driver_opts()
    for backend in configured_backends:
        for opt_list in driver_opts:
            if configured_backends[backend] not in opt_list:
                continue

            LOG.debug("Registering options for group %s", backend)
            conf.register_opts(driver_opts[opt_list], group=backend)


def _load_multi_store(conf, store_entry,
                      invoke_load=True,
                      backend=None):
    if backend:
        invoke_args = [conf, backend]
    else:
        invoke_args = [conf]
    try:
        LOG.debug("Attempting to import store %s", store_entry)
        mgr = driver.DriverManager('glance_store.drivers',
                                   store_entry,
                                   invoke_args=invoke_args,
                                   invoke_on_load=invoke_load)
        return mgr.driver
    except RuntimeError as e:
        LOG.warning("Failed to load driver %(driver)s. The "
                    "driver will be disabled", dict(driver=str([driver, e])))


def _load_multi_stores(conf, reserved_stores=None):
    enabled_backends = conf.enabled_backends
    if reserved_stores:
        enabled_backends.update(reserved_stores)
        _RESERVED_STORES.update(reserved_stores)

    for backend, store_entry in enabled_backends.items():
        try:
            # FIXME(flaper87): Don't hide BadStoreConfiguration
            # exceptions. These exceptions should be propagated
            # to the user of the library.
            store_instance = _load_multi_store(conf, store_entry,
                                               backend=backend)

            if not store_instance:
                continue

            yield (store_entry, store_instance, backend)

        except exceptions.BadStoreConfiguration:
            continue


def create_multi_stores(conf=CONF, reserved_stores=None):
    """
    Registers all store modules and all schemes from the given configuration
    object.

    :param conf: A oslo_config (or compatible) object
    :param reserved_stores: A list of stores for the consuming service's
                            internal use.  The list must be the same
                            format as the ``enabled_backends`` configuration
                            setting.  The default value is None
    :return: The number of stores configured
    :raises: ``glance_store.exceptions.BackendException``

    *Configuring Multiple Backends*

    The backends to be configured are expected to be found in the
    ``enabled_backends`` configuration variable in the DEFAULT group
    of the object.  The format for the variable is a dictionary of
    key:value pairs where the key is an arbitrary store identifier
    and the value is the store type identifier for the store.

    The type identifiers must be defined in the  ``[entry points]``
    section of the glance_store ``setup.cfg`` file as values for
    the ``glance_store.drivers`` configuration.  (See the default
    ``setup.cfg`` file for an example.)  The store type identifiers
    for the currently supported drivers are already defined in the file.

    Thus an example value for ``enabled_backends`` is::

        {'store_one': 'http', 'store_two': 'file', 'store_three': 'rbd'}

    The ``reserved_stores`` parameter, if included, must have the same
    format.  There is no difference between the ``enabled_backends`` and
    ``reserved_stores`` from the glance_store point of view: the reserved
    stores are a convenience for the consuming service, which may wish
    to handle the two sets of stores differently.

    *The Default Store*

    If you wish to set a default store, its store identifier should be
    defined as the value of the ``default_backend`` configuration option
    in the ``glance_store`` group of the ``conf`` parameter.  The store
    identifier, or course, should be specified as one of the keys in the
    ``enabled_backends`` dict.  It is recommended that a default store
    be set.

    *Configuring Individual Backends*

    To configure each store mentioned in the ``enabled_backends``
    configuration option, you must define an option group with the
    same name as the store identifier.  The options defined for that
    backend will depend upon the store type; consult the documentation
    for the appropriate backend driver to determine what these are.

    For example, given the ``enabled_backends`` example above, you
    would put the following in the configuration file that loads the
    ``conf`` object::

        [DEFAULT]
        enabled_backends = store_one:rbd,store_two:file,store_three:http

        [store_one]
        store_description = "A human-readable string aimed at end users"
        rbd_store_chunk_size = 8
        rbd_store_pool = images
        rbd_store_user = admin
        rbd_store_ceph_conf = /etc/ceph/ceph.conf

        [store_two]
        store_description = "Human-readable description of this store"
        filesystem_store_datadir = /opt/stack/data/glance/store_two

        [store_three]
        store_description = "A read-only store"
        https_ca_certificates_file = /opt/stack/certs/gs.cert

        [glance_store]
        default_backend = store_two

    The ``store_description`` options may be used by a consuming service.
    As recommended above, this file also defines a default backend.
    """

    store_count = 0
    scheme_map = {}
    for (store_entry, store_instance,
         store_identifier) in _load_multi_stores(
            conf, reserved_stores=reserved_stores):
        try:
            schemes = store_instance.get_schemes()
            store_instance.configure(re_raise_bsc=False)
        except NotImplementedError:
            continue

        if not schemes:
            raise exceptions.BackendException(
                _('Unable to register store %s. No schemes associated '
                  'with it.') % store_entry)
        else:
            LOG.debug("Registering store %s with schemes %s",
                      store_entry, schemes)

            loc_cls = store_instance.get_store_location_class()
            for scheme in schemes:
                if scheme not in scheme_map:
                    scheme_map[scheme] = {}
                scheme_map[scheme][store_identifier] = {
                    'store': store_instance,
                    'location_class': loc_cls,
                    'store_entry': store_entry
                }
                location.register_scheme_backend_map(scheme_map)
                store_count += 1

    return store_count


def verify_store():
    store_id = CONF.glance_store.default_backend
    if not store_id:
        msg = _("'default_backend' config option is not set.")
        raise RuntimeError(msg)

    try:
        get_store_from_store_identifier(store_id)
    except exceptions.UnknownScheme:
        msg = _("Store for identifier %s not found") % store_id
        raise RuntimeError(msg)


def get_store_weight(store_identifier):
    """Determine backing store weightage from identifier.

    Given a store identifier, return the appropriate weight of store
    from memory.
    """
    enabled_backends = CONF.enabled_backends
    enabled_backends.update(_RESERVED_STORES)

    try:
        scheme = enabled_backends[store_identifier]
    except KeyError:
        msg = _("Store for identifier %s not found") % store_identifier
        raise exceptions.UnknownScheme(msg)

    try:
        backend_map = location.SCHEME_TO_CLS_BACKEND_MAP[scheme]
        scheme_info = backend_map[store_identifier]
    except KeyError:
        raise exceptions.UnknownScheme(scheme=scheme)

    store = scheme_info['store']
    if store:
        return store.weight

    return 0


def get_store_from_store_identifier(store_identifier):
    """Determine backing store from identifier.

    Given a store identifier, return the appropriate store object
    for handling that scheme.
    """
    scheme_map = {}
    enabled_backends = CONF.enabled_backends
    enabled_backends.update(_RESERVED_STORES)

    try:
        scheme = enabled_backends[store_identifier]
    except KeyError:
        msg = _("Store for identifier %s not found") % store_identifier
        raise exceptions.UnknownScheme(msg)

    if scheme not in location.SCHEME_TO_CLS_BACKEND_MAP:
        raise exceptions.UnknownScheme(scheme=scheme)

    scheme_info = location.SCHEME_TO_CLS_BACKEND_MAP[scheme][store_identifier]
    store = scheme_info['store']

    if not store.is_capable(capabilities.BitMasks.DRIVER_REUSABLE):
        # Driver instance isn't stateless so it can't
        # be reused safely and need recreation.
        store_entry = scheme_info['store_entry']
        store = _load_multi_store(store.conf, store_entry, invoke_load=True,
                                  backend=store_identifier)
        store.configure()
        try:
            loc_cls = store.get_store_location_class()
            for new_scheme in store.get_schemes():
                if new_scheme not in scheme_map:
                    scheme_map[new_scheme] = {}

                scheme_map[new_scheme][store_identifier] = {
                    'store': store,
                    'location_class': loc_cls,
                    'store_entry': store_entry
                }
                location.register_scheme_backend_map(scheme_map)
        except NotImplementedError:
            scheme_info['store'] = store

    return store


def add(conf, image_id, data, size, backend, context=None,
        verifier=None):
    if not backend:
        backend = conf.glance_store.default_backend

    store = get_store_from_store_identifier(backend)
    return store_add_to_backend(image_id, data, size, store, context,
                                verifier)


def add_with_multihash(conf, image_id, data, size, backend, hashing_algo,
                       scheme=None, context=None, verifier=None):
    if not backend:
        backend = conf.glance_store.default_backend

    store = get_store_from_store_identifier(backend)
    return store_add_to_backend_with_multihash(
        image_id, data, size, hashing_algo, store, context, verifier)


def _check_metadata(store, metadata):
    if not isinstance(metadata, dict):
        msg = (_("The storage driver %(driver)s returned invalid "
                 " metadata %(metadata)s. This must be a dictionary type")
               % dict(driver=str(store), metadata=str(metadata)))
        LOG.error(msg)
        raise exceptions.BackendException(msg)
    try:
        check_location_metadata(metadata)
    except exceptions.BackendException as e:
        e_msg = (_("A bad metadata structure was returned from the "
                   "%(driver)s storage driver: %(metadata)s.  %(e)s.") %
                 dict(driver=store,
                      metadata=metadata,
                      e=e))
        LOG.error(e_msg)
        raise exceptions.BackendException(e_msg)


def store_add_to_backend(image_id, data, size, store, context=None,
                         verifier=None):
    """A wrapper around a call to each stores add() method.

    This gives glance a common place to check the output.

    :param image_id:  The image add to which data is added
    :param data: The data to be stored
    :param size: The length of the data in bytes
    :param store: The store to which the data is being added
    :param context: The request context
    :param verifier: An object used to verify signatures for images
    :param backend: Name of the backend to store the image
    :return: The url location of the file,
             the size amount of data,
             the checksum of the data
             the storage systems metadata dictionary for the location
    """
    (location, size, checksum, metadata) = store.add(image_id,
                                                     data,
                                                     size,
                                                     context=context,
                                                     verifier=verifier)

    if metadata is not None:
        _check_metadata(store, metadata)

    return (location, size, checksum, metadata)


def store_add_to_backend_with_multihash(
        image_id, data, size, hashing_algo, store,
        context=None, verifier=None):
    """
    A wrapper around a call to each store's add() method that requires
    a hashing_algo identifier and returns a 5-tuple including the
    "multihash" computed using the specified hashing_algo.  (This
    is an enhanced version of store_add_to_backend(), which is left
    as-is for backward compatibility.)

    :param image_id:  The image add to which data is added
    :param data: The data to be stored
    :param size: The length of the data in bytes
    :param store: The store to which the data is being added
    :param hashing_algo: A hashlib algorithm identifier (string)
    :param context: The request context
    :param verifier: An object used to verify signatures for images
    :return: The url location of the file,
             the size amount of data,
             the checksum of the data,
             the multihash of the data,
             the storage system's metadata dictionary for the location
    :raises: ``glance_store.exceptions.BackendException``
             ``glance_store.exceptions.UnknownHashingAlgo``
    """

    if hashing_algo not in hashlib.algorithms_available:
        raise exceptions.UnknownHashingAlgo(algo=hashing_algo)

    (location, size, checksum, multihash, metadata) = store.add(
        image_id, data, size, hashing_algo, context=context, verifier=verifier)

    if metadata is not None:
        _check_metadata(store, metadata)

    return (location, size, checksum, multihash, metadata)


def check_location_metadata(val, key=''):
    if isinstance(val, dict):
        for key in val:
            check_location_metadata(val[key], key=key)
    elif isinstance(val, list):
        ndx = 0
        for v in val:
            check_location_metadata(v, key='%s[%d]' % (key, ndx))
            ndx = ndx + 1
    elif not isinstance(val, str):
        raise exceptions.BackendException(_("The image metadata key %(key)s "
                                            "has an invalid type of %(type)s. "
                                            "Only dict, list, and unicode are "
                                            "supported.")
                                          % dict(key=key, type=type(val)))


def delete(uri, backend, context=None):
    """Removes chunks of data from backend specified by uri."""
    if backend:
        loc = location.get_location_from_uri_and_backend(
            uri, backend, conf=CONF)
        store = get_store_from_store_identifier(backend)
        return store.delete(loc, context=context)

    LOG.warning('Backend is not set to image, searching all backends based on '
                'location URI.')

    backends = CONF.enabled_backends
    for backend in backends:
        try:
            if not uri.startswith(backends[backend]):
                continue

            loc = location.get_location_from_uri_and_backend(
                uri, backend, conf=CONF)
            store = get_store_from_store_identifier(backend)
            return store.delete(loc, context=context)
        except (exceptions.NotFound, exceptions.UnknownScheme):
            continue

    raise exceptions.NotFound(_("Image not found in any configured backend"))


def set_acls_for_multi_store(location_uri, backend, public=False,
                             read_tenants=[],
                             write_tenants=None, context=None):

    if write_tenants is None:
        write_tenants = []

    loc = location.get_location_from_uri_and_backend(
        location_uri, backend, conf=CONF)
    store = get_store_from_store_identifier(backend)
    try:
        store.set_acls(loc, public=public,
                       read_tenants=read_tenants,
                       write_tenants=write_tenants,
                       context=context)
    except NotImplementedError:
        LOG.debug("Skipping store.set_acls... not implemented")


def get(uri, backend, offset=0, chunk_size=None, context=None):
    """Yields chunks of data from backend specified by uri."""

    if backend:
        loc = location.get_location_from_uri_and_backend(uri, backend,
                                                         conf=CONF)
        store = get_store_from_store_identifier(backend)

        return store.get(loc, offset=offset,
                         chunk_size=chunk_size,
                         context=context)

    LOG.warning('Backend is not set to image, searching all backends based on '
                'location URI.')

    backends = CONF.enabled_backends
    for backend in backends:
        try:
            if not uri.startswith(backends[backend]):
                continue

            loc = location.get_location_from_uri_and_backend(
                uri, backend, conf=CONF)
            store = get_store_from_store_identifier(backend)
            data, size = store.get(loc, offset=offset,
                                   chunk_size=chunk_size,
                                   context=context)
            if data:
                return data, size
        except (exceptions.NotFound, exceptions.UnknownScheme):
            continue

    raise exceptions.NotFound(_("Image not found in any configured backend"))


def get_known_schemes_for_multi_store():
    """Returns list of known schemes."""
    return location.SCHEME_TO_CLS_BACKEND_MAP.keys()


def get_size_from_uri_and_backend(uri, backend, context=None):
    """Retrieves image size from backend specified by uri."""

    loc = location.get_location_from_uri_and_backend(
        uri, backend, conf=CONF)
    store = get_store_from_store_identifier(backend)
    return store.get_size(loc, context=context)



# Copyright 2011 OpenStack Foundation
# Copyright 2012 RedHat Inc.
# Copyright 2018 Verizon Wireless
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Base class for all storage backends"""

from functools import wraps
import logging

from oslo_config import cfg
from oslo_utils import importutils
from oslo_utils import units

from glance_store import capabilities
from glance_store import exceptions
from glance_store.i18n import _

LOG = logging.getLogger(__name__)

CONF = cfg.CONF
SHARED_CONF_GROUP = 'backend_defaults'

_MULTI_BACKEND_OPTS = [
    cfg.StrOpt('store_description',
               help=_("""
This option will be used to provide a constructive information about
the store backend to end users. Using /v2/stores-info call user can
seek more information on all available backends.

""")),
    cfg.IntOpt('weight',
               help=_("""
This option is used to define a relative weight for this store over
any others that are configured. The actual value of the weight is meaningless
and only serves to provide a "sort order" compared to others. Any stores
with the same weight will be treated as equivalent.
"""),
               default=0),
]


class Store(capabilities.StoreCapability):

    OPTIONS = None
    MULTI_BACKEND_OPTIONS = _MULTI_BACKEND_OPTS
    READ_CHUNKSIZE = 4 * units.Mi  # 4M
    WRITE_CHUNKSIZE = READ_CHUNKSIZE

    def __init__(self, conf, backend=None):
        """
        Initialize the Store
        """

        super(Store, self).__init__()

        self.conf = conf
        self.backend_group = backend
        self.store_location_class = None
        self._url_prefix = None

        try:
            if self.OPTIONS is not None:
                group = 'glance_store'
                if self.backend_group:
                    group = self.backend_group
                    if self.MULTI_BACKEND_OPTIONS is not None:
                        self.conf.register_opts(
                            self.MULTI_BACKEND_OPTIONS, group=group)

                self.conf.register_opts(self.OPTIONS, group=group)
                self.conf.register_opts(self.OPTIONS, group=SHARED_CONF_GROUP)
        except cfg.DuplicateOptError:
            pass

    @property
    def url_prefix(self):
        return self._url_prefix

    @property
    def weight(self):
        if self.backend_group is None:
            # NOTE(danms): A backend with no config group can not have a
            # weight set, so just return the default
            return 0
        else:
            return getattr(self.conf, self.backend_group).weight

    def configure(self, re_raise_bsc=False):
        """
        Configure the store to use the stored configuration options
        and initialize capabilities based on current configuration.

        Any store that needs special configuration should implement
        this method.
        """

        try:
            self.configure_add()
        except exceptions.BadStoreConfiguration as e:
            self.unset_capabilities(capabilities.BitMasks.WRITE_ACCESS)
            msg = _("Failed to configure store correctly: %s "
                    "Disabling add method.") % e
            LOG.warning(msg)
            if re_raise_bsc:
                raise
        finally:
            self.update_capabilities()

    def get_schemes(self):
        """
        Returns a tuple of schemes which this store can handle.
        """
        raise NotImplementedError

    def get_store_location_class(self):
        """
        Returns the store location class that is used by this store.
        """
        if not self.store_location_class:
            class_name = "%s.StoreLocation" % (self.__module__)
            LOG.debug("Late loading location class %s", class_name)
            self.store_location_class = importutils.import_class(class_name)
        return self.store_location_class

    def configure_add(self):
        """
        This is like `configure` except that it's specifically for
        configuring the store to accept objects.

        If the store was not able to successfully configure
        itself, it should raise `exceptions.BadStoreConfiguration`.
        """
        # NOTE(flaper87): This should probably go away

    @capabilities.check
    def get(self, location, offset=0, chunk_size=None, context=None):
        """
        Takes a `glance_store.location.Location` object that indicates
        where to find the image file, and returns a tuple of generator
        (for reading the image file) and image_size

        :param location: `glance_store.location.Location` object, supplied
                        from glance_store.location.get_location_from_uri()
        :raises: `glance.exceptions.NotFound` if image does not exist
        """
        raise NotImplementedError

    def get_size(self, location, context=None):
        """
        Takes a `glance_store.location.Location` object that indicates
        where to find the image file, and returns the size

        :param location: `glance_store.location.Location` object, supplied
                        from glance_store.location.get_location_from_uri()
        :raises: `glance_store.exceptions.NotFound` if image does not exist
        """
        raise NotImplementedError

    # NOTE(rosmaita): use the @glance_store.driver.back_compat_add
    # annotation on implementions for backward compatibility with
    # pre-0.26.0 add().  Need backcompat because pre-0.26.0 returned
    # a 4 tuple, this returns a 5-tuple
    @capabilities.check
    def add(self, image_id, image_file, image_size, hashing_algo,
            context=None, verifier=None):
        """
        Stores an image file with supplied identifier to the backend
        storage system and returns a tuple containing information
        about the stored image.

        :param image_id: The opaque image identifier
        :param image_file: The image data to write, as a file-like object
        :param image_size: The size of the image data to write, in bytes
        :param hashing_algo: A hashlib algorithm identifier (string)
        :param context: A context object
        :param verifier: An object used to verify signatures for images

        :returns: tuple of: (1) URL in backing store, (2) bytes written,
                  (3) checksum, (4) multihash value, and (5) a dictionary
                  with storage system specific information
        :raises: `glance_store.exceptions.Duplicate` if the image already
                 exists
        """
        raise NotImplementedError

    @capabilities.check
    def delete(self, location, context=None):
        """
        Takes a `glance_store.location.Location` object that indicates
        where to find the image file to delete

        :param location: `glance_store.location.Location` object, supplied
                  from glance_store.location.get_location_from_uri()
        :raises: `glance_store.exceptions.NotFound` if image does not exist
        """
        raise NotImplementedError

    def set_acls(self, location, public=False, read_tenants=None,
                 write_tenants=None, context=None):
        """
        Sets the read and write access control list for an image in the
        backend store.

        :param location: `glance_store.location.Location` object, supplied
                  from glance_store.location.get_location_from_uri()
        :param public: A boolean indicating whether the image should be public.
        :param read_tenants: A list of tenant strings which should be granted
                      read access for an image.
        :param write_tenants: A list of tenant strings which should be granted
                      write access for an image.
        """
        raise NotImplementedError


def back_compat_add(store_add_fun):
    """
    Provides backward compatibility for the 0.26.0+ Store.add() function.
    In 0.26.0, the 'hashing_algo' parameter is introduced and Store.add()
    returns a 5-tuple containing a computed 'multihash' value.

    This wrapper behaves as follows:

    If no hashing_algo identifier is supplied as an argument, the response
    is the pre-0.26.0 4-tuple of::

    (backend_url, bytes_written, checksum, metadata_dict)

    If a hashing_algo is supplied, the response is a 5-tuple::

    (backend_url, bytes_written, checksum, multihash, metadata_dict)

    The wrapper detects the presence of a 'hashing_algo' argument both
    by examining named arguments and positionally.
    """

    @wraps(store_add_fun)
    def add_adapter(*args, **kwargs):
        """
        Wrapper for the store 'add' function.  If no hashing_algo identifier
        is supplied, the response is the pre-0.25.0 4-tuple of::

        (backend_url, bytes_written, checksum, metadata_dict)

        If a hashing_algo is supplied, the response is a 5-tuple::

        (backend_url, bytes_written, checksum, multihash, metadata_dict)
        """
        # strategy: assume this until we determine otherwise
        back_compat_required = True

        # specify info about 0.26.0 Store.add() call (can't introspect
        # this because the add method is wrapped by the capabilities
        # check)
        p_algo = 4
        max_args = 7

        num_args = len(args)
        num_kwargs = len(kwargs)

        if num_args + num_kwargs == max_args:
            # everything is present, including hashing_algo
            back_compat_required = False
        elif ('hashing_algo' in kwargs or
              (num_args >= p_algo + 1 and isinstance(args[p_algo], str))):
            # there is a hashing_algo argument present
            back_compat_required = False
        else:
            # this is a pre-0.26.0-style call, so let's figure out
            # whether to insert the hashing_algo in the args or kwargs
            if kwargs and 'image_' in ''.join(kwargs):
                # if any of the image_* is named, everything after it
                # must be named as well, so slap the algo into kwargs
                kwargs['hashing_algo'] = 'md5'
            else:
                args = args[:p_algo] + ('md5',) + args[p_algo:]

        # business time
        (backend_url,
         bytes_written,
         checksum,
         multihash,
         metadata_dict) = store_add_fun(*args, **kwargs)

        if back_compat_required:
            return (backend_url, bytes_written, checksum, metadata_dict)

        return (backend_url, bytes_written, checksum, multihash,
                metadata_dict)

    return add_adapter


class BackendGroupConfiguration(object):

    def __init__(self, store_opts, config_group=None, conf=None):
        """Initialize configuration.

        This takes care of grafting the implementation's config
        values into the config group and shared defaults. We will try to
        pull values from the specified 'config_group', but fall back to
        defaults from the SHARED_CONF_GROUP.
        """
        self.config_group = config_group
        self.conf = conf or CONF

        # set the local conf so that __call__'s know what to use
        self._ensure_config_values(store_opts)
        self.backend_conf = self.conf._get(self.config_group)
        self.shared_backend_conf = self.conf._get(SHARED_CONF_GROUP)

    def _safe_register(self, opt, group):
        try:
            CONF.register_opt(opt, group=group)
        except cfg.DuplicateOptError:
            pass  # If it's already registered ignore it

    def _ensure_config_values(self, store_opts):
        """Register the options in the shared group.

        When we go to get a config option we will try the backend specific
        group first and fall back to the shared group. We override the default
        from all the config options for the backend group so we can know if it
        was set or not.
        """
        for opt in store_opts:
            self._safe_register(opt, SHARED_CONF_GROUP)
            # Assuming they aren't the same groups, graft on the options into
            # the backend group and override its default value.
            if self.config_group != SHARED_CONF_GROUP:
                self._safe_register(opt, self.config_group)
                self.conf.set_default(opt.name, None, group=self.config_group)

    def append_config_values(self, store_opts):
        self._ensure_config_values(store_opts)

    def set_default(self, opt_name, default):
        self.conf.set_default(opt_name, default, group=SHARED_CONF_GROUP)

    def get(self, key, default=None):
        return getattr(self, key, default)

    def __getattr__(self, opt_name):
        # Don't use self.X to avoid reentrant call to __getattr__()
        backend_conf = object.__getattribute__(self, 'backend_conf')
        opt_value = getattr(backend_conf, opt_name)
        if opt_value is None:
            shared_conf = object.__getattribute__(self, 'shared_backend_conf')
            opt_value = getattr(shared_conf, opt_name)
        return opt_value



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from glance_store import exceptions
from glance_store import location
from glance_store.tests import base


class TestStoreLocation(base.StoreBaseTest):

    def setUp(self):
        super(TestStoreLocation, self).setUp()

    def test_scheme_validation(self):
        valid_schemas = ("file://", "http://")
        correct_uri = "file://test"
        location.StoreLocation.validate_schemas(correct_uri, valid_schemas)
        incorrect_uri = "fake://test"
        self.assertRaises(exceptions.BadStoreUri,
                          location.StoreLocation.validate_schemas,
                          incorrect_uri, valid_schemas)



# Copyright 2025 RedHat Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import builtins
import errno
import hashlib
import io
import os
import stat
from unittest import mock
import uuid

import fixtures
from oslo_utils import units

from glance_store._drivers import filesystem
from glance_store import exceptions
from glance_store import location


class TestFilerStoreBase(object):

    def _test_get(self):
        image_id = str(uuid.uuid4())
        file_contents = b"chunk00000remainder"
        image_file = io.BytesIO(file_contents)

        if self.multistore:
            loc, size, checksum, metadata = self.store.add(
                image_id, image_file, len(file_contents))
            self.assertEqual(self.backend, metadata['store'])
        else:
            loc, size, checksum, multihash, _ = self.store.add(
                image_id, image_file, len(file_contents), self.hash_algo)

        # Now read it back...
        uri = "file:///%s/%s" % (self.test_dir, image_id)
        if self.multistore:
            loc = location.get_location_from_uri_and_backend(uri, self.backend,
                                                             conf=self.conf)
        else:
            loc = location.get_location_from_uri(uri, conf=self.conf)

        (image_file, image_size) = self.store.get(loc)

        data_chunks = list(image_file)
        data = b"".join(data_chunks)

        self.assertEqual(file_contents, data)
        self.assertEqual(2, len(data_chunks))

    def _test_get_random_access(self):
        image_id = str(uuid.uuid4())
        file_contents = b"chunk00000remainder"
        image_file = io.BytesIO(file_contents)

        if self.multistore:
            loc, size, checksum, metadata = self.store.add(
                image_id, image_file, len(file_contents))
            self.assertEqual(self.backend, metadata['store'])
        else:
            loc, size, checksum, multihash, _ = self.store.add(
                image_id, image_file, len(file_contents), self.hash_algo)

        # Now read it back...
        uri = "file:///%s/%s" % (self.test_dir, image_id)
        if self.multistore:
            loc = location.get_location_from_uri_and_backend(uri, self.backend,
                                                             conf=self.conf)
        else:
            loc = location.get_location_from_uri(uri, conf=self.conf)

        # Test reading one byte at a time from every offset
        data = bytearray()
        for offset in range(len(file_contents)):
            image_file, image_size = self.store.get(
                loc, offset=offset, chunk_size=1)
            for chunk in image_file:
                data.extend(chunk)
        self.assertEqual(file_contents, bytes(data))

        # Test reading a chunk from a specific offset
        chunk_size = 5
        image_file, image_size = self.store.get(
            loc, offset=chunk_size, chunk_size=chunk_size)
        chunk_data = b"".join(chunk for chunk in image_file)
        self.assertEqual(b'00000', chunk_data)
        self.assertEqual(chunk_size, image_size)

    def _test_get_non_existing(self):
        image_id = str(uuid.uuid4())
        uri = "file:///%s/%s" % (self.test_dir, image_id)
        if self.multistore:
            loc = location.get_location_from_uri_and_backend(uri, self.backend,
                                                             conf=self.conf)
        else:
            loc = location.get_location_from_uri(uri, conf=self.conf)

        self.assertRaises(exceptions.NotFound, self.store.get, loc)

    def _test_add(self):
        filesystem.ChunkedFile.CHUNKSIZE = units.Ki
        expected_image_id = str(uuid.uuid4())
        expected_file_size = 5 * units.Ki  # 5K
        expected_file_contents = b"*" * expected_file_size
        expected_checksum = hashlib.md5(expected_file_contents,
                                        usedforsecurity=False).hexdigest()
        expected_location = "file://%s/%s" % (self.test_dir,
                                              expected_image_id)
        image_file = io.BytesIO(expected_file_contents)

        if self.multistore:
            loc, size, checksum, metadata = self.store.add(
                expected_image_id, image_file, expected_file_size)
            self.assertEqual(self.backend, metadata['store'])
        else:
            loc, size, checksum, multihash, _ = self.store.add(
                expected_image_id, image_file, expected_file_size,
                self.hash_algo)

        self.assertEqual(expected_location, loc)
        self.assertEqual(expected_file_size, size)
        self.assertEqual(expected_checksum, checksum)

        # Check the location
        uri = "file:///%s/%s" % (self.test_dir, expected_image_id)
        if self.multistore:
            loc = location.get_location_from_uri_and_backend(uri, self.backend,
                                                             conf=self.conf)
        else:
            loc = location.get_location_from_uri(uri, conf=self.conf)

        (new_image_file, new_image_size) = self.store.get(loc)
        new_image_contents = b""
        new_image_file_size = 0

        for chunk in new_image_file:
            new_image_file_size += len(chunk)
            new_image_contents += chunk

        self.assertEqual(expected_file_contents, new_image_contents)
        self.assertEqual(expected_file_size, new_image_file_size)

    def _test_add_image_exceeding_max_size_raises_exception(self):
        expected_image_id = str(uuid.uuid4())
        path = os.path.join(self.test_dir, expected_image_id)
        # expected total size
        expected_file_size = 1020
        # simulate input with extra data
        image_file = io.BytesIO(b'a' * (expected_file_size + 100))

        # Call method and assert exception
        if self.multistore:
            self.assertRaisesRegex(
                exceptions.Invalid, "Size exceeds: expected",
                self.store.add, expected_image_id, image_file,
                expected_file_size)
        else:
            self.assertRaisesRegex(
                exceptions.Invalid, "Size exceeds: expected",
                self.store.add, expected_image_id, image_file,
                expected_file_size, self.hash_algo)

        # Verify partial data is deleted from backend
        self.assertFalse(os.path.exists(path))

        # Verify that the stream's position reflects the number of bytes read,
        # which should be exactly at expected_file_size plus the last buffer
        # size read.
        expected_read = expected_file_size + self.store.WRITE_CHUNKSIZE
        self.assertEqual(expected_read, image_file.tell(),
                         "The stream was not read only up to the expected "
                         "size.")

    def _test_write_less_than_declared_raises_exception(self):
        # Setup
        expected_image_id = str(uuid.uuid4())
        path = os.path.join(self.test_dir, expected_image_id)
        # expected total size
        actual_data_size = 800
        # declared size larger than actual data
        declared_size = 1000
        image_file = io.BytesIO(b'b' * actual_data_size)

        # Call method and assert exception
        if self.multistore:
            self.assertRaisesRegex(
                exceptions.Invalid, "Size mismatch: expected",
                self.store.add, expected_image_id, image_file,
                declared_size)
        else:
            self.assertRaisesRegex(
                exceptions.Invalid, "Size mismatch: expected",
                self.store.add, expected_image_id, image_file,
                declared_size, self.hash_algo)

        # Verify partial data is deleted from backend
        self.assertFalse(os.path.exists(path))
        # The input buffer should be fully read
        self.assertEqual(actual_data_size, image_file.tell(),
                         "Input stream was not fully read as expected")

    def _do_test_add(self, enable_thin_provisoning):
        """Test that we can add an image via the filesystem backend."""
        self.config(filesystem_store_chunk_size=units.Ki,
                    filesystem_thin_provisioning=enable_thin_provisoning,
                    group=self.backend)
        self.store.configure()

        filesystem.ChunkedFile.CHUNKSIZE = units.Ki
        expected_image_id = str(uuid.uuid4())
        expected_file_size = 5 * units.Ki  # 5K
        expected_file_contents = b"*" * expected_file_size
        expected_checksum = hashlib.md5(expected_file_contents,
                                        usedforsecurity=False).hexdigest()
        expected_multihash = hashlib.sha256(expected_file_contents).hexdigest()
        expected_location = "file://%s/%s" % (self.test_dir,
                                              expected_image_id)
        image_file = io.BytesIO(expected_file_contents)

        if self.multistore:
            loc, size, checksum, metadata = self.store.add(
                expected_image_id, image_file, expected_file_size)
            self.assertEqual({'store': self.backend}, metadata)
        else:
            loc, size, checksum, multihash, _ = self.store.add(
                expected_image_id, image_file, expected_file_size,
                self.hash_algo)
            self.assertEqual(expected_multihash, multihash)

        self.assertEqual(expected_location, loc)
        self.assertEqual(expected_file_size, size)
        self.assertEqual(expected_checksum, checksum)

        uri = "file:///%s/%s" % (self.test_dir, expected_image_id)
        if self.multistore:
            loc = location.get_location_from_uri_and_backend(uri, self.backend,
                                                             conf=self.conf)
        else:
            loc = location.get_location_from_uri(uri, conf=self.conf)

        (new_image_file, new_image_size) = self.store.get(loc)
        new_image_contents = b""
        new_image_file_size = 0

        for chunk in new_image_file:
            new_image_file_size += len(chunk)
            new_image_contents += chunk

        self.assertEqual(expected_file_contents, new_image_contents)
        self.assertEqual(expected_file_size, new_image_file_size)

    def _do_test_thin_provisioning(self, content, size, truncate, write, thin):
        self.config(filesystem_store_chunk_size=units.Ki,
                    filesystem_thin_provisioning=thin,
                    group=self.backend)
        self.store.configure()

        image_file = io.BytesIO(content)
        image_id = str(uuid.uuid4())
        with mock.patch.object(builtins, 'open') as popen:
            if self.multistore:
                self.store.add(image_id, image_file, size)
            else:
                self.store.add(image_id, image_file, size, self.hash_algo)

            write_count = popen.return_value.__enter__().write.call_count
            truncate_count = popen.return_value.__enter__().truncate.call_count
            self.assertEqual(write_count, write)
            self.assertEqual(truncate_count, truncate)

    def _test_add_with_verifier(self):
        """Test that 'verifier.update' is called when verifier is provided."""
        verifier = mock.MagicMock(name='mock_verifier')
        self.config(filesystem_store_chunk_size=units.Ki,
                    group=self.backend)
        self.store.configure()

        image_id = str(uuid.uuid4())
        file_size = units.Ki  # 1K
        file_contents = b"*" * file_size
        image_file = io.BytesIO(file_contents)

        if self.multistore:
            location, size, checksum, metadata = self.store.add(
                image_id, image_file, file_size, verifier=verifier)
        else:
            self.store.add(image_id, image_file, file_size, self.hash_algo,
                           verifier=verifier)

        verifier.update.assert_called_with(file_contents)

    def _test_add_check_metadata_with_invalid_mountpoint_location(self):
        in_metadata = [{'id': 'abcdefg',
                        'mountpoint': '/xyz/images'}]
        if self.multistore:
            location, size, checksum, metadata = self._store_image(in_metadata)
            self.assertEqual({'store': 'file1'}, metadata)
        else:
            location, size, checksum, multihash, metadata = self._store_image(
                in_metadata)
            self.assertEqual({}, metadata)

    def _test_add_check_metadata_list_with_invalid_mountpoint_locations(self):
        in_metadata = [{'id': 'abcdefg', 'mountpoint': '/xyz/images'},
                       {'id': 'xyz1234', 'mountpoint': '/pqr/images'}]
        if self.multistore:
            location, size, checksum, metadata = self._store_image(in_metadata)
            self.assertEqual({'store': 'file1'}, metadata)
        else:
            location, size, checksum, multihash, metadata = self._store_image(
                in_metadata)
            self.assertEqual({}, metadata)

    def _test_add_check_metadata_list_with_valid_mountpoint_locations(self):
        in_metadata = [{'id': 'abcdefg', 'mountpoint': '/tmp'},
                       {'id': 'xyz1234', 'mountpoint': '/xyz'}]
        if self.multistore:
            location, size, checksum, metadata = self._store_image(in_metadata)
            self.assertEqual("file1", metadata["store"])
        else:
            location, size, checksum, multihash, metadata = self._store_image(
                in_metadata)

        self.assertEqual(in_metadata[0], metadata)

    def _test_add_check_metadata_bad_nosuch_file(self):
        expected_image_id = str(uuid.uuid4())
        jsonfilename = os.path.join(self.test_dir,
                                    "storage_metadata.%s" % expected_image_id)

        self.config(filesystem_store_metadata_file=jsonfilename,
                    group=self.backend)
        expected_file_size = 10
        expected_file_contents = b"*" * expected_file_size
        image_file = io.BytesIO(expected_file_contents)

        if self.multistore:
            location, size, checksum, metadata = self.store.add(
                expected_image_id, image_file, expected_file_size)
            self.assertEqual({'store': self.backend}, metadata)
        else:
            location, size, checksum, multihash, metadata = self.store.add(
                expected_image_id, image_file, expected_file_size,
                self.hash_algo)
            self.assertEqual({}, metadata)

    def _test_add_already_existing(self):
        """
        Tests that adding an image with an existing identifier
        raises an appropriate exception
        """
        filesystem.ChunkedFile.CHUNKSIZE = units.Ki
        image_id = str(uuid.uuid4())
        file_size = 5 * units.Ki  # 5K
        file_contents = b"*" * file_size
        image_file = io.BytesIO(file_contents)
        dup_image_file = io.BytesIO(b"nevergonnamakeit")
        if self.multistore:
            location, size, checksum, metadata = self.store.add(
                image_id, image_file, file_size)
            self.assertEqual("file1", metadata["store"])
            self.assertRaises(exceptions.Duplicate,
                              self.store.add,
                              image_id, dup_image_file, 0)
        else:
            location, size, checksum, multihash, metadata = self.store.add(
                image_id, image_file, file_size, self.hash_algo)
            self.assertEqual({}, metadata)
            self.assertRaises(exceptions.Duplicate,
                              self.store.add,
                              image_id, dup_image_file, 0, self.hash_algo)

    def _do_test_add_write_failure(self, errno, exception):
        filesystem.ChunkedFile.CHUNKSIZE = units.Ki
        image_id = str(uuid.uuid4())
        file_size = 5 * units.Ki  # 5K
        file_contents = b"*" * file_size
        path = os.path.join(self.test_dir, image_id)
        image_file = io.BytesIO(file_contents)

        with mock.patch.object(builtins, 'open') as popen:
            e = IOError()
            e.errno = errno
            popen.side_effect = e

            if self.multistore:
                self.assertRaises(exception,
                                  self.store.add,
                                  image_id, image_file, 0)
            else:
                self.assertRaises(exception,
                                  self.store.add,
                                  image_id, image_file, 0, self.hash_algo)
            self.assertFalse(os.path.exists(path))

    def _test_add_cleanup_on_read_failure(self):
        """Tests partial image is cleaned up after a read failure."""
        filesystem.ChunkedFile.CHUNKSIZE = units.Ki
        image_id = str(uuid.uuid4())
        file_size = 5 * units.Ki  # 5K
        file_contents = b"*" * file_size
        path = os.path.join(self.test_dir, image_id)
        image_file = io.BytesIO(file_contents)

        def fake_Error(size):
            raise AttributeError()

        with mock.patch.object(image_file, 'read') as mock_read:
            mock_read.side_effect = fake_Error

            if self.multistore:
                self.assertRaises(AttributeError,
                                  self.store.add,
                                  image_id, image_file, 0)
            else:
                self.assertRaises(AttributeError,
                                  self.store.add,
                                  image_id, image_file, 0, self.hash_algo)
            self.assertFalse(os.path.exists(path))

    def _test_delete(self):
        """
        Test we can delete an existing image in the filesystem store
        """
        # First add an image
        image_id = str(uuid.uuid4())
        file_size = 5 * units.Ki  # 5K
        file_contents = b"*" * file_size
        image_file = io.BytesIO(file_contents)

        if self.multistore:
            loc, size, checksum, metadata = self.store.add(
                image_id, image_file, file_size)
            self.assertEqual('file1', metadata['store'])
        else:
            loc, size, checksum, multihash, _ = self.store.add(
                image_id, image_file, file_size, self.hash_algo)

        # Now check that we can delete it
        uri = "file:///%s/%s" % (self.test_dir, image_id)
        if self.multistore:
            loc = location.get_location_from_uri_and_backend(uri, 'file1',
                                                             conf=self.conf)
        else:
            loc = location.get_location_from_uri(uri, conf=self.conf)

        self.store.delete(loc)
        self.assertRaises(exceptions.NotFound, self.store.get, loc)

    def _test_delete_non_existing(self):
        """
        Test that trying to delete a file that doesn't exist
        raises an error
        """
        if self.multistore:
            loc = location.get_location_from_uri_and_backend(
                "file:///tmp/glance-tests/non-existing", 'file1',
                conf=self.conf)
        else:
            loc = location.get_location_from_uri(
                "file:///tmp/glance-tests/non-existing", conf=self.conf)

        self.assertRaises(exceptions.NotFound,
                          self.store.delete,
                          loc)

    def _test_delete_forbidden(self):
        """
        Tests that trying to delete a file without permissions
        raises the correct error
        """
        # First add an image
        image_id = str(uuid.uuid4())
        file_size = 5 * units.Ki  # 5K
        file_contents = b"*" * file_size
        image_file = io.BytesIO(file_contents)

        if self.multistore:
            loc, size, checksum, metadata = self.store.add(
                image_id, image_file, file_size)
        else:
            loc, size, checksum, multihash, _ = self.store.add(
                image_id, image_file, file_size, self.hash_algo)

        uri = "file:///%s/%s" % (self.test_dir, image_id)
        if self.multistore:
            loc = location.get_location_from_uri_and_backend(uri, 'file1',
                                                             conf=self.conf)
        else:
            loc = location.get_location_from_uri(uri, conf=self.conf)

        # Mock unlink to raise an OSError for lack of permissions
        # and make sure we can't delete the image
        with mock.patch.object(os, 'unlink') as unlink:
            e = OSError()
            e.errno = errno.EACCES
            unlink.side_effect = e

            self.assertRaises(exceptions.Forbidden,
                              self.store.delete,
                              loc)

            # Make sure the image didn't get deleted
            self.store.get(loc)

    def _test_configure_add_with_multi_datadirs(self):
        """Test multiple filesystems are parsed correctly."""
        store_map = [self.useFixture(fixtures.TempDir()).path,
                     self.useFixture(fixtures.TempDir()).path]
        self.conf.set_override('filesystem_store_datadir',
                               override=None,
                               group=self.backend)
        self.conf.set_override('filesystem_store_datadir',
                               override=None,
                               group='backend_defaults')
        self.conf.set_override('filesystem_store_datadirs',
                               [store_map[0] + ":100",
                                store_map[1] + ":200"],
                               group=self.backend)
        self.store.configure_add()

        expected_priority_map = {100: [store_map[0]], 200: [store_map[1]]}
        expected_priority_list = [200, 100]
        self.assertEqual(expected_priority_map, self.store.priority_data_map)
        self.assertEqual(expected_priority_list, self.store.priority_list)

    def _test_configure_add_with_metadata_file_success(self):
        metadata = {'id': 'asdf1234',
                    'mountpoint': '/tmp'}
        self._create_metadata_json_file(metadata, group=self.backend)
        self.store.configure_add()
        self.assertEqual([metadata], self.store.FILESYSTEM_STORE_METADATA)

    def _test_configure_add_check_metadata_list_of_dicts_success(self):
        metadata = [{'id': 'abcdefg', 'mountpoint': '/xyz/images'},
                    {'id': 'xyz1234', 'mountpoint': '/tmp/'}]
        self._create_metadata_json_file(metadata, group=self.backend)
        self.store.configure_add()
        self.assertEqual(metadata, self.store.FILESYSTEM_STORE_METADATA)

    def _test_configure_add_check_metadata_success_list_val_for_some_key(self):
        metadata = {'akey': ['value1', 'value2'], 'id': 'asdf1234',
                    'mountpoint': '/tmp'}
        self._create_metadata_json_file(metadata, group=self.backend)
        self.store.configure_add()
        self.assertEqual([metadata], self.store.FILESYSTEM_STORE_METADATA)

    def _test_configure_add_check_metadata_bad_data(self):
        metadata = {'akey': 10, 'id': 'asdf1234',
                    'mountpoint': '/tmp'}  # only unicode is allowed
        self._create_metadata_json_file(metadata, group=self.backend)
        self.assertRaises(exceptions.BadStoreConfiguration,
                          self.store.configure_add)

    def _test_configure_add_check_metadata_with_no_id_or_mountpoint(self):
        metadata = {'mountpoint': '/tmp'}
        self._create_metadata_json_file(metadata, group=self.backend)
        self.assertRaises(exceptions.BadStoreConfiguration,
                          self.store.configure_add)

        metadata = {'id': 'asdfg1234'}
        self._create_metadata_json_file(metadata, group=self.backend)
        self.assertRaises(exceptions.BadStoreConfiguration,
                          self.store.configure_add)

    def _test_cfg_add_check_metadata_id_or_mountpoint_is_not_string(self):
        metadata = {'id': 10, 'mountpoint': '/tmp'}
        self._create_metadata_json_file(metadata, group=self.backend)
        self.assertRaises(exceptions.BadStoreConfiguration,
                          self.store.configure_add)

        metadata = {'id': 'asdf1234', 'mountpoint': 12345}
        self._create_metadata_json_file(metadata, group=self.backend)
        self.assertRaises(exceptions.BadStoreConfiguration,
                          self.store.configure_add)

    def _test_cfg_add_check_metadata_list_with_no_id_or_mountpoint(self):
        metadata = [{'id': 'abcdefg', 'mountpoint': '/xyz/images'},
                    {'mountpoint': '/pqr/images'}]
        self._create_metadata_json_file(metadata, group=self.backend)
        self.assertRaises(exceptions.BadStoreConfiguration,
                          self.store.configure_add)

        metadata = [{'id': 'abcdefg'},
                    {'id': 'xyz1234', 'mountpoint': '/pqr/images'}]
        self._create_metadata_json_file(metadata, group=self.backend)
        self.assertRaises(exceptions.BadStoreConfiguration,
                          self.store.configure_add)

    def _test_add_check_metadata_list_id_or_mountpoint_is_not_string(self):
        metadata = [{'id': 'abcdefg', 'mountpoint': '/xyz/images'},
                    {'id': 1234, 'mountpoint': '/pqr/images'}]
        self._create_metadata_json_file(metadata, group=self.backend)
        self.assertRaises(exceptions.BadStoreConfiguration,
                          self.store.configure_add)

        metadata = [{'id': 'abcdefg', 'mountpoint': 1234},
                    {'id': 'xyz1234', 'mountpoint': '/pqr/images'}]
        self._create_metadata_json_file(metadata, group=self.backend)
        self.assertRaises(exceptions.BadStoreConfiguration,
                          self.store.configure_add)

    def _test_configure_add_same_dir_multiple_times(self):
        """Tests handling of same dir in config multiple times.

        Tests BadStoreConfiguration exception is raised if same directory
        is specified multiple times in filesystem_store_datadirs with different
        priorities.
        """
        store_map = [self.useFixture(fixtures.TempDir()).path,
                     self.useFixture(fixtures.TempDir()).path]
        self.conf.clear_override('filesystem_store_datadir',
                                 group=self.backend)
        self.conf.set_override('filesystem_store_datadirs',
                               [store_map[0] + ":100",
                                store_map[1] + ":200",
                                store_map[0] + ":300"],
                               group=self.backend)
        self.assertRaises(exceptions.BadStoreConfiguration,
                          self.store.configure_add)

    def _test_configure_add_same_dir_multiple_times_same_priority(self):
        """Tests handling of same dir in config multiple times.

        Tests BadStoreConfiguration exception is raised if same directory
        is specified multiple times in filesystem_store_datadirs with the same
        priority.
        """
        store_map = [self.useFixture(fixtures.TempDir()).path,
                     self.useFixture(fixtures.TempDir()).path]
        self.conf.set_override('filesystem_store_datadir',
                               override=None,
                               group=self.backend)
        self.conf.set_override('filesystem_store_datadir',
                               override=None,
                               group='backend_defaults')
        self.conf.set_override('filesystem_store_datadirs',
                               [store_map[0] + ":100",
                                store_map[1] + ":200",
                                store_map[0] + ":100"],
                               group=self.backend)
        try:
            self.store.configure()
        except exceptions.BadStoreConfiguration:
            self.fail("configure() raised BadStoreConfiguration unexpectedly!")

        # Test that we can add an image via the filesystem backend
        filesystem.ChunkedFile.CHUNKSIZE = 1024
        expected_image_id = str(uuid.uuid4())
        expected_file_size = 5 * units.Ki  # 5K
        expected_file_contents = b"*" * expected_file_size
        expected_checksum = hashlib.md5(expected_file_contents,
                                        usedforsecurity=False).hexdigest()
        expected_location = "file://%s/%s" % (store_map[1],
                                              expected_image_id)
        image_file = io.BytesIO(expected_file_contents)

        if self.multistore:
            loc, size, checksum, metadata = self.store.add(expected_image_id,
                                                           image_file,
                                                           expected_file_size)
            self.assertEqual("file1", metadata["store"])
        else:
            loc, size, checksum, multihash, _ = self.store.add(
                expected_image_id, image_file, expected_file_size,
                self.hash_algo)

        self.assertEqual(expected_location, loc)
        self.assertEqual(expected_file_size, size)
        self.assertEqual(expected_checksum, checksum)

        if self.multistore:
            loc = location.get_location_from_uri_and_backend(
                expected_location, "file1", conf=self.conf)
        else:
            loc = location.get_location_from_uri(expected_location,
                                                 conf=self.conf)

        (new_image_file, new_image_size) = self.store.get(loc)
        new_image_contents = b""
        new_image_file_size = 0

        for chunk in new_image_file:
            new_image_file_size += len(chunk)
            new_image_contents += chunk

        self.assertEqual(expected_file_contents, new_image_contents)
        self.assertEqual(expected_file_size, new_image_file_size)

    def _test_add_with_multiple_dirs(self):
        """Test adding multiple filesystem directories."""
        store_map = [self.useFixture(fixtures.TempDir()).path,
                     self.useFixture(fixtures.TempDir()).path]
        self.conf.set_override('filesystem_store_datadir',
                               override=None,
                               group=self.backend)
        self.conf.set_override('filesystem_store_datadir',
                               override=None,
                               group='backend_defaults')

        self.conf.set_override('filesystem_store_datadirs',
                               [store_map[0] + ":100",
                                store_map[1] + ":200"],
                               group=self.backend)

        self.store.configure()

        # Test that we can add an image via the filesystem backend
        filesystem.ChunkedFile.CHUNKSIZE = units.Ki
        expected_image_id = str(uuid.uuid4())
        expected_file_size = 5 * units.Ki  # 5K
        expected_file_contents = b"*" * expected_file_size
        expected_checksum = hashlib.md5(expected_file_contents,
                                        usedforsecurity=False).hexdigest()
        expected_multihash = hashlib.sha256(expected_file_contents).hexdigest()
        expected_location = "file://%s/%s" % (store_map[1],
                                              expected_image_id)
        image_file = io.BytesIO(expected_file_contents)

        if self.multistore:
            loc, size, checksum, metadata = self.store.add(
                expected_image_id, image_file, expected_file_size)
            self.assertEqual("file1", metadata["store"])
        else:
            loc, size, checksum, multihash, _ = self.store.add(
                expected_image_id, image_file, expected_file_size,
                self.hash_algo)
            self.assertEqual(expected_multihash, multihash)

        self.assertEqual(expected_location, loc)
        self.assertEqual(expected_file_size, size)
        self.assertEqual(expected_checksum, checksum)

        if self.multistore:
            loc = location.get_location_from_uri_and_backend(
                expected_location, self.backend, conf=self.conf)
        else:
            loc = location.get_location_from_uri(expected_location,
                                                 conf=self.conf)
        (new_image_file, new_image_size) = self.store.get(loc)
        new_image_contents = b""
        new_image_file_size = 0

        for chunk in new_image_file:
            new_image_file_size += len(chunk)
            new_image_contents += chunk

        self.assertEqual(expected_file_contents, new_image_contents)
        self.assertEqual(expected_file_size, new_image_file_size)

    def _test_add_with_multiple_dirs_storage_full(self):
        """Tests adding dirs with storage full.

        Test StorageFull exception is raised if no filesystem directory
        is found that can store an image.
        """
        store_map = [self.useFixture(fixtures.TempDir()).path,
                     self.useFixture(fixtures.TempDir()).path]
        self.conf.set_override('filesystem_store_datadir',
                               override=None,
                               group=self.backend)
        self.conf.set_override('filesystem_store_datadir',
                               override=None,
                               group='backend_defaults')
        self.conf.set_override('filesystem_store_datadirs',
                               [store_map[0] + ":100",
                                store_map[1] + ":200"],
                               group=self.backend)

        self.store.configure_add()

        def fake_get_capacity_info(mount_point):
            return 0

        with mock.patch.object(self.store, '_get_capacity_info') as capacity:
            capacity.return_value = 0

            filesystem.ChunkedFile.CHUNKSIZE = units.Ki
            expected_image_id = str(uuid.uuid4())
            expected_file_size = 5 * units.Ki  # 5K
            expected_file_contents = b"*" * expected_file_size
            image_file = io.BytesIO(expected_file_contents)

            if self.multistore:
                self.assertRaises(exceptions.StorageFull, self.store.add,
                                  expected_image_id, image_file,
                                  expected_file_size)
            else:
                self.assertRaises(exceptions.StorageFull, self.store.add,
                                  expected_image_id, image_file,
                                  expected_file_size, self.hash_algo)

    def _test_configure_add_with_file_perm(self):
        """Tests adding with permissions.

        Tests filesystem specified by filesystem_store_file_perm
        are parsed correctly.
        """
        store = self.useFixture(fixtures.TempDir()).path
        self.conf.set_override('filesystem_store_datadir', store,
                               group=self.backend)
        self.conf.set_override('filesystem_store_file_perm', 700,  # -rwx------
                               group=self.backend)
        self.store.configure_add()
        self.assertEqual(self.store.datadir, store)

    def _test_configure_add_with_inaccessible_file_perm(self):
        """Tests adding with inaccessible file permissions.

        Tests BadStoreConfiguration exception is raised if an invalid
        file permission specified in filesystem_store_file_perm.
        """
        store = self.useFixture(fixtures.TempDir()).path
        self.conf.set_override('filesystem_store_datadir', store,
                               group=self.backend)
        self.conf.set_override('filesystem_store_file_perm', 7,  # -------rwx
                               group=self.backend)
        self.assertRaises(exceptions.BadStoreConfiguration,
                          self.store.configure_add)

    def _test_add_with_file_perm_for_group_other_users_access(self):
        """Tests adding image with file permissions.

        Test that we can add an image via the filesystem backend with a
        required image file permission.
        """
        store = self.useFixture(fixtures.TempDir()).path
        self.conf.set_override('filesystem_store_datadir', store,
                               group=self.backend)
        self.conf.set_override('filesystem_store_file_perm', 744,  # -rwxr--r--
                               group=self.backend)

        # -rwx------
        os.chmod(store, 0o700)
        self.assertEqual(0o700, stat.S_IMODE(os.stat(store)[stat.ST_MODE]))

        self.store.configure_add()

        filesystem.Store.WRITE_CHUNKSIZE = units.Ki
        expected_image_id = str(uuid.uuid4())
        expected_file_size = 5 * units.Ki  # 5K
        expected_file_contents = b"*" * expected_file_size
        expected_checksum = hashlib.md5(expected_file_contents,
                                        usedforsecurity=False).hexdigest()
        expected_location = "file://%s/%s" % (store,
                                              expected_image_id)
        image_file = io.BytesIO(expected_file_contents)

        if self.multistore:
            location, size, checksum, metadata = self.store.add(
                expected_image_id, image_file, expected_file_size)
            self.assertEqual("file1", metadata["store"])
        else:
            location, size, checksum, multihash, metadata = self.store.add(
                expected_image_id, image_file, expected_file_size,
                self.hash_algo)

        self.assertEqual(expected_location, location)
        self.assertEqual(expected_file_size, size)
        self.assertEqual(expected_checksum, checksum)

        # -rwx--x--x for store directory
        self.assertEqual(0o711, stat.S_IMODE(os.stat(store)[stat.ST_MODE]))
        # -rwxr--r-- for image file
        mode = os.stat(expected_location[len('file:/'):])[stat.ST_MODE]
        perm = int(str(getattr(self.conf,
                               self.backend).filesystem_store_file_perm), 8)
        self.assertEqual(perm, stat.S_IMODE(mode))

    def _test_add_with_file_perm_for_owner_users_access(self):
        """Tests adding image with file permissions.

        Test that we can add an image via the filesystem backend with a
        required image file permission.
        """
        store = self.useFixture(fixtures.TempDir()).path
        self.conf.set_override('filesystem_store_datadir', store,
                               group=self.backend)
        self.conf.set_override('filesystem_store_file_perm', 600,  # -rw-------
                               group=self.backend)

        # -rwx------
        os.chmod(store, 0o700)
        self.assertEqual(0o700, stat.S_IMODE(os.stat(store)[stat.ST_MODE]))

        self.store.configure_add()

        filesystem.Store.WRITE_CHUNKSIZE = units.Ki
        expected_image_id = str(uuid.uuid4())
        expected_file_size = 5 * units.Ki  # 5K
        expected_file_contents = b"*" * expected_file_size
        expected_checksum = hashlib.md5(expected_file_contents,
                                        usedforsecurity=False).hexdigest()
        expected_location = "file://%s/%s" % (store,
                                              expected_image_id)
        image_file = io.BytesIO(expected_file_contents)

        if self.multistore:
            location, size, checksum, metadata = self.store.add(
                expected_image_id, image_file, expected_file_size)
            self.assertEqual("file1", metadata["store"])
        else:
            location, size, checksum, multihash, metadata = self.store.add(
                expected_image_id, image_file, expected_file_size,
                self.hash_algo)

        self.assertEqual(expected_location, location)
        self.assertEqual(expected_file_size, size)
        self.assertEqual(expected_checksum, checksum)

        # -rwx------ for store directory
        self.assertEqual(0o700, stat.S_IMODE(os.stat(store)[stat.ST_MODE]))
        # -rw------- for image file
        mode = os.stat(expected_location[len('file:/'):])[stat.ST_MODE]
        perm = int(str(getattr(self.conf,
                               self.backend).filesystem_store_file_perm), 8)
        self.assertEqual(perm, stat.S_IMODE(mode))

    def _test_configure_add_chunk_size(self):
        # This definitely won't be the default
        chunk_size = units.Gi
        self.config(filesystem_store_chunk_size=chunk_size,
                    group=self.backend)
        self.store.configure_add()

        self.assertEqual(chunk_size, self.store.chunk_size)
        self.assertEqual(chunk_size, self.store.READ_CHUNKSIZE)
        self.assertEqual(chunk_size, self.store.WRITE_CHUNKSIZE)



# Copyright 2014 OpenStack, LLC
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests the VMware Datastore backend store"""

from unittest import mock
import uuid

from oslo_utils import units
from oslo_vmware.objects import datacenter as oslo_datacenter
from oslo_vmware.objects import datastore as oslo_datastore

import glance_store._drivers.vmware_datastore as vm_store
from glance_store import backend
from glance_store.tests import base
from glance_store.tests.unit import test_store_capabilities
from glance_store.tests.unit import test_vmware_store_base as vmware_base


FAKE_UUID = str(uuid.uuid4())

FIVE_KB = 5 * units.Ki

VMWARE_DS = {
    'debug': True,
    'known_stores': ['vmware_datastore'],
    'default_store': 'vsphere',
    'vmware_server_host': '127.0.0.1',
    'vmware_server_username': 'username',
    'vmware_server_password': 'password',
    'vmware_store_image_dir': '/openstack_glance',
    'vmware_insecure': 'True',
    'vmware_datastores': ['a:b:0'],
}


def format_location(host_ip, folder_name, image_id, datastores):
    """
    Helper method that returns a VMware Datastore store URI given
    the component pieces.
    """
    scheme = 'vsphere'
    (datacenter_path, datastore_name, weight) = datastores[0].split(':')
    return ("%s://%s/folder%s/%s?dcPath=%s&dsName=%s"
            % (scheme, host_ip, folder_name,
               image_id, datacenter_path, datastore_name))


def fake_datastore_obj(*args, **kwargs):
    dc_obj = oslo_datacenter.Datacenter(ref='fake-ref',
                                        name='fake-name')
    dc_obj.path = args[0]
    return oslo_datastore.Datastore(ref='fake-ref',
                                    datacenter=dc_obj,
                                    name=args[1])


class TestStore(base.StoreBaseTest,
                vmware_base.TestVMwareStoreBase,
                test_store_capabilities.TestStoreCapabilitiesChecking):

    @mock.patch.object(vm_store.Store, '_get_datastore')
    @mock.patch('oslo_vmware.api.VMwareAPISession')
    def setUp(self, mock_api_session, mock_get_datastore):
        """Establish a clean test environment."""
        super(TestStore, self).setUp()

        # Set class attributes for multistore and backend
        self.multistore = False
        self.backend = 'glance_store'

        vm_store.Store.CHUNKSIZE = 2
        default_store = VMWARE_DS['default_store']
        self.config(default_store=default_store, stores=['vmware'])
        backend.register_opts(self.conf)
        self.config(group='glance_store',
                    vmware_server_username='admin',
                    vmware_server_password='admin',
                    vmware_server_host=VMWARE_DS['vmware_server_host'],
                    vmware_insecure=VMWARE_DS['vmware_insecure'],
                    vmware_datastores=VMWARE_DS['vmware_datastores'])

        mock_get_datastore.side_effect = fake_datastore_obj
        backend.create_stores(self.conf)

        self.store = backend.get_store_from_scheme('vsphere')

        self.store.store_image_dir = (
            VMWARE_DS['vmware_store_image_dir'])

        self.hash_algo = 'sha256'

    def _mock_http_connection(self):
        return mock.patch('http.client.HTTPConnection')

    def test_get(self):
        """Test a "normal" retrieval of an image in chunks."""
        self._test_get()

    def test_get_non_existing(self):
        """
        Test that trying to retrieve an image that doesn't exist
        raises an error
        """
        self._test_get_non_existing()

    def test_add(self):
        """Test that we can add an image via the VMware backend."""
        self._test_add()

    def test_add_size_zero(self):
        """
        Test that when specifying size zero for the image to add,
        the actual size of the image is returned.
        """
        self._test_add_size_zero()

    def test_add_with_verifier(self):
        """Test that the verifier is passed to the _Reader during add."""
        self._test_add_with_verifier()

    def test_add_with_verifier_size_zero(self):
        """Test that the verifier is passed to the _ChunkReader during add."""
        self._test_add_with_verifier_size_zero()

    def test_delete(self):
        """Test we can delete an existing image in the VMware store."""
        self._test_delete()

    def test_delete_non_existing(self):
        """
        Test that trying to delete an image that doesn't exist raises an error
        """
        self._test_delete_non_existing()

    def test_get_size(self):
        """
        Test we can get the size of an existing image in the VMware store
        """
        self._test_get_size()

    def test_get_size_non_existing(self):
        """
        Test that trying to retrieve an image size that doesn't exist
        raises an error
        """
        self._test_get_size_non_existing()

    def test_reader_full(self):
        """Test the reader reads full content correctly."""
        self._test_reader_full()

    def test_reader_partial(self):
        """Test the reader reads partial content correctly."""
        self._test_reader_partial()

    def test_reader_with_verifier(self):
        """Test the reader works with verifier."""
        self._test_reader_with_verifier()

    def test_sanity_check_api_retry_count(self):
        """Test that sanity check raises if api_retry_count is <= 0."""
        self._test_sanity_check_api_retry_count()

    def test_sanity_check_task_poll_interval(self):
        """Test that sanity check raises if task_poll_interval is <= 0."""
        self._test_sanity_check_task_poll_interval()

    def test_sanity_check_multiple_datastores(self):
        """Test sanity check with multiple datastores."""
        self._test_sanity_check_multiple_datastores()

    def test_parse_datastore_info_and_weight_less_opts(self):
        """Test parsing datastore with less options."""
        self._test_parse_datastore_info_and_weight_less_opts()

    def test_parse_datastore_info_and_weight_invalid_weight(self):
        """Test parsing datastore with invalid weight."""
        self._test_parse_datastore_info_and_weight_invalid_weight()

    def test_parse_datastore_info_and_weight_empty_opts(self):
        """Test parsing datastore with empty options."""
        self._test_parse_datastore_info_and_weight_empty_opts()

    def test_parse_datastore_info_and_weight(self):
        """Test parsing datastore with valid options."""
        self._test_parse_datastore_info_and_weight()

    def test_parse_datastore_info_and_weight_default_weight(self):
        """Test parsing datastore with default weight."""
        self._test_parse_datastore_info_and_weight_default_weight()

    def test_unexpected_status(self):
        """Test handling of unexpected HTTP status."""
        self._test_unexpected_status()

    def test_unexpected_status_no_response_body(self):
        """Test handling of unexpected HTTP status with no response body."""
        self._test_unexpected_status_no_response_body()

    def test_reset_session(self):
        """Test resetting the session."""
        self._test_reset_session()

    def test_build_vim_cookie_header_active(self):
        """Test building VIM cookie header with active session."""
        self._test_build_vim_cookie_header_active()

    def test_build_vim_cookie_header_expired(self):
        """Test building VIM cookie header with expired session."""
        self._test_build_vim_cookie_header_expired()

    def test_build_vim_cookie_header_expired_noverify(self):
        """Test building VIM cookie header with expired session no verify."""
        self._test_build_vim_cookie_header_expired_noverify()

    def test_add_ioerror(self):
        """Test handling of IOError during add."""
        self._test_add_ioerror()

    def test_qs_sort_with_literal_question_mark(self):
        """Test URL sorting with literal question mark."""
        self._test_qs_sort_with_literal_question_mark()

    def test_build_datastore_weighted_map(self):
        """Test building datastore weighted map."""
        self._test_build_datastore_weighted_map()

    def test_build_datastore_weighted_map_equal_weight(self):
        """Test building datastore weighted map with equal weights."""
        self._test_build_datastore_weighted_map_equal_weight()

    def test_build_datastore_weighted_map_empty_list(self):
        """Test building datastore weighted map with empty list."""
        self._test_build_datastore_weighted_map_empty_list()

    def test_select_datastore_insufficient_freespace(self):
        """Test selecting datastore with insufficient free space."""
        self._test_select_datastore_insufficient_freespace()

    def test_select_datastore_insufficient_fs_one_ds(self):
        """Test selecting datastore with less free space on one datastore."""
        self._test_select_datastore_insufficient_fs_one_ds()

    def test_select_datastore_equal_freespace(self):
        """Test selecting datastore with equal free space."""
        self._test_select_datastore_equal_freespace()

    def test_select_datastore_contention(self):
        """Test selecting datastore with contention."""
        self._test_select_datastore_contention()

    def test_select_datastore_empty_list(self):
        """Test selecting datastore with empty list."""
        self._test_select_datastore_empty_list()

    def test_get_datacenter_ref(self):
        """Test getting datacenter reference."""
        self._test_get_datacenter_ref()

    def test_http_get_redirect(self):
        """Test HTTP GET with redirects."""
        self._test_http_get_redirect()

    def test_http_get_max_redirects(self):
        """Test HTTP GET with max redirects exceeded."""
        self._test_http_get_max_redirects()

    def test_http_get_redirect_invalid(self):
        """Test HTTP GET with invalid redirect."""
        self._test_http_get_redirect_invalid()



# Copyright 2018 RedHat Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests the filesystem backend store"""

import errno
import io
import json
import os
import time
import uuid

import fixtures
import futurist
from oslo_config import cfg
from oslo_utils import units
from unittest import mock

import glance_store as store
from glance_store._drivers import filesystem
from glance_store import exceptions
from glance_store import location
from glance_store.tests import base
from glance_store.tests.unit import test_filesystem_store_base as file_base
from glance_store.tests.unit import test_store_capabilities


class TestMultiStore(base.MultiStoreBaseTest,
                     file_base.TestFilerStoreBase,
                     test_store_capabilities.TestStoreCapabilitiesChecking):
    # NOTE(flaper87): temporary until we
    # can move to a fully-local lib.
    # (Swift store's fault)
    _CONF = cfg.ConfigOpts()

    def setUp(self):
        """Establish a clean test environment."""
        super(TestMultiStore, self).setUp()
        # Set default values for multistore and backend
        self.multistore = True
        self.backend = 'file1'

        self.enabled_backends = {
            "file1": "file",
            "file2": "file",
        }
        self.conf = self._CONF
        self.conf(args=[])
        self.conf.register_opt(cfg.DictOpt('enabled_backends'))
        self.config(enabled_backends=self.enabled_backends)
        store.register_store_opts(self.conf)
        self.config(default_backend='file1', group='glance_store')

        # Ensure stores + locations cleared
        location.SCHEME_TO_CLS_BACKEND_MAP = {}
        store.create_multi_stores(self.conf)

        self.addCleanup(setattr, location, 'SCHEME_TO_CLS_BACKEND_MAP',
                        dict())
        self.test_dir = self.useFixture(fixtures.TempDir()).path
        self.addCleanup(self.conf.reset)

        self.store = filesystem.Store(self.conf, backend='file1')
        self.config(filesystem_store_datadir=self.test_dir,
                    filesystem_store_chunk_size=10,
                    group="file1")
        self.store.configure()
        self.register_store_backend_schemes(self.store, 'file', 'file1')

    def _create_metadata_json_file(self, metadata, group=None):
        expected_image_id = str(uuid.uuid4())
        jsonfilename = os.path.join(self.test_dir,
                                    "storage_metadata.%s" % expected_image_id)

        self.config(filesystem_store_metadata_file=jsonfilename,
                    group=group)
        with open(jsonfilename, 'w') as fptr:
            json.dump(metadata, fptr)

    def _store_image(self, in_metadata):
        expected_image_id = str(uuid.uuid4())
        expected_file_size = 10
        expected_file_contents = b"*" * expected_file_size
        image_file = io.BytesIO(expected_file_contents)
        self.store.FILESYSTEM_STORE_METADATA = in_metadata
        return self.store.add(expected_image_id, image_file,
                              expected_file_size)

    def test_location_url_prefix_is_set(self):
        expected_url_prefix = "file://%s" % self.test_dir
        self.assertEqual(expected_url_prefix, self.store.url_prefix)

    def test_get(self):
        """Test a "normal" retrieval of an image in chunks."""
        self._test_get()

    def test_get_random_access(self):
        """Test a "normal" retrieval of an image in chunks."""
        # First add an image...
        self._test_get_random_access()

    def test_get_non_existing(self):
        """Test trying to retrieve a file that doesn't exist raises error."""
        self._test_get_non_existing()

    def test_get_non_existing_identifier(self):
        """Test trying to retrieve a store that doesn't exist raises error."""
        self.assertRaises(exceptions.UnknownScheme,
                          location.get_location_from_uri_and_backend,
                          "file:///%s/non-existing" % self.test_dir,
                          'file3', conf=self.conf)

    def test_add(self):
        """Test that we can add an image via the filesystem backend."""
        self._test_add()

    def test_add_to_different_backned(self):
        """Test that we can add an image via the filesystem backend."""
        # Temporarily change backend for this test
        original_backend = self.backend
        self.backend = 'file2'

        self.store = filesystem.Store(self.conf, backend='file2')
        self.config(filesystem_store_datadir=self.test_dir,
                    group="file2")
        self.store.configure()
        self.register_store_backend_schemes(self.store, 'file', 'file2')
        self._test_add()

        # Restore original backend
        self.backend = original_backend

    def test_add_image_exceeding_max_size_raises_exception(self):
        self._test_add_image_exceeding_max_size_raises_exception()

    def test_write_less_than_declared_raises_exception(self):
        self._test_write_less_than_declared_raises_exception()

    def test_thin_provisioning_is_disabled_by_default(self):
        self.assertEqual(self.store.thin_provisioning, False)

    def test_add_with_thick_provisioning(self):
        self._do_test_add(enable_thin_provisoning=False)

    def test_add_with_thin_provisioning(self):
        self._do_test_add(enable_thin_provisoning=True)

    def test_add_thick_provisioning_with_holes_in_file(self):
        """
        Tests that a file which contains null bytes chunks is fully
        written with a thick provisioning configuration.
        """
        chunk_size = units.Ki  # 1K
        content = b"*" * chunk_size + b"\x00" * chunk_size + b"*" * chunk_size
        self._do_test_thin_provisioning(content, 3 * chunk_size, 0, 3, False)

    def test_add_thin_provisioning_with_holes_in_file(self):
        """
        Tests that a file which contains null bytes chunks is sparsified
        with a thin provisioning configuration.
        """
        chunk_size = units.Ki  # 1K
        content = b"*" * chunk_size + b"\x00" * chunk_size + b"*" * chunk_size
        self._do_test_thin_provisioning(content, 3 * chunk_size, 1, 2, True)

    def test_add_thick_provisioning_without_holes_in_file(self):
        """
        Tests that a file which not contain null bytes chunks is fully
        written with a thick provisioning configuration.
        """
        chunk_size = units.Ki  # 1K
        content = b"*" * 3 * chunk_size
        self._do_test_thin_provisioning(content, 3 * chunk_size, 0, 3, False)

    def test_add_thin_provisioning_without_holes_in_file(self):
        """
        Tests that a file which not contain null bytes chunks is fully
        written with a thin provisioning configuration.
        """
        chunk_size = units.Ki  # 1K
        content = b"*" * 3 * chunk_size
        self._do_test_thin_provisioning(content, 3 * chunk_size, 0, 3, True)

    def test_add_thick_provisioning_with_partial_holes_in_file(self):
        """
        Tests that a file which contains null bytes not aligned with
        chunk size is fully written with a thick provisioning configuration.
        """
        chunk_size = units.Ki  # 1K
        my_chunk = int(chunk_size * 1.5)
        content = b"*" * my_chunk + b"\x00" * my_chunk + b"*" * my_chunk
        self._do_test_thin_provisioning(content, 3 * my_chunk, 0, 5, False)

    def test_add_thin_provisioning_with_partial_holes_in_file(self):
        """
        Tests that a file which contains null bytes not aligned with
        chunk size is sparsified with a thin provisioning configuration.
        """
        chunk_size = units.Ki  # 1K
        my_chunk = int(chunk_size * 1.5)
        content = b"*" * my_chunk + b"\x00" * my_chunk + b"*" * my_chunk
        self._do_test_thin_provisioning(content, 3 * my_chunk, 1, 4, True)

    def test_add_with_verifier(self):
        self._test_add_with_verifier()

    def test_add_check_metadata_with_invalid_mountpoint_location(self):
        self._test_add_check_metadata_with_invalid_mountpoint_location()

    def test_add_check_metadata_list_with_invalid_mountpoint_locations(self):
        self._test_add_check_metadata_list_with_invalid_mountpoint_locations()

    def test_add_check_metadata_list_with_valid_mountpoint_locations(self):
        self._test_add_check_metadata_list_with_valid_mountpoint_locations()

    def test_add_check_metadata_bad_nosuch_file(self):
        self._test_add_check_metadata_bad_nosuch_file()

    def test_add_already_existing(self):
        self._test_add_already_existing()

    def test_add_storage_full(self):
        """Tests adding an image without enough space.

        Tests that adding an image without enough space on disk
        raises an appropriate exception.
        """
        self._do_test_add_write_failure(
            errno.ENOSPC, exceptions.StorageFull)

    def test_add_file_too_big(self):
        """Tests adding a very large image.

        Tests that adding an excessively large image file
        raises an appropriate exception.
        """
        self._do_test_add_write_failure(
            errno.EFBIG, exceptions.StorageFull)

    def test_add_storage_write_denied(self):
        """Tests adding an image without store permissions.

        Tests that adding an image with insufficient filestore permissions
        raises an appropriate exception.
        """
        self._do_test_add_write_failure(errno.EACCES,
                                        exceptions.StorageWriteDenied)

    def test_add_other_failure(self):
        """Tests other IOErrors do not raise a StorageFull exception."""
        self._do_test_add_write_failure(
            errno.ENOTDIR, IOError)

    def test_add_cleanup_on_read_failure(self):
        self._test_add_cleanup_on_read_failure()

    def test_delete(self):
        self._test_delete()

    def test_delete_non_existing(self):
        self._test_delete_non_existing()

    def test_delete_forbidden(self):
        self._test_delete_forbidden()

    def test_configure_add_with_multi_datadirs(self):
        self._test_configure_add_with_multi_datadirs()

    def test_configure_add_with_metadata_file_success(self):
        self._test_configure_add_with_metadata_file_success()

    def test_configure_add_check_metadata_list_of_dicts_success(self):
        self._test_configure_add_check_metadata_list_of_dicts_success()

    def test_configure_add_check_metadata_success_list_val_for_some_key(self):
        self._test_configure_add_check_metadata_success_list_val_for_some_key()

    def test_configure_add_check_metadata_bad_data(self):
        self._test_configure_add_check_metadata_bad_data()

    def test_configure_add_check_metadata_with_no_id_or_mountpoint(self):
        self._test_configure_add_check_metadata_with_no_id_or_mountpoint()

    def test_configure_add_check_metadata_id_or_mountpoint_is_not_string(self):
        self._test_cfg_add_check_metadata_id_or_mountpoint_is_not_string()

    def test_configure_add_check_metadata_list_with_no_id_or_mountpoint(self):
        self._test_cfg_add_check_metadata_list_with_no_id_or_mountpoint()

    def test_add_check_metadata_list_id_or_mountpoint_is_not_string(self):
        self._test_add_check_metadata_list_id_or_mountpoint_is_not_string()

    def test_configure_add_same_dir_multiple_times(self):
        self._test_configure_add_same_dir_multiple_times()

    def test_configure_add_same_dir_multiple_times_same_priority(self):
        self._test_configure_add_same_dir_multiple_times_same_priority()

    def test_add_with_multiple_dirs(self):
        self._test_add_with_multiple_dirs()

    def test_add_with_multiple_dirs_storage_full(self):
        self._test_add_with_multiple_dirs_storage_full()

    def test_configure_add_with_file_perm(self):
        self._test_configure_add_with_file_perm()

    def test_configure_add_with_inaccessible_file_perm(self):
        self._test_configure_add_with_inaccessible_file_perm()

    def test_add_with_file_perm_for_group_other_users_access(self):
        self._test_add_with_file_perm_for_group_other_users_access()

    def test_add_with_file_perm_for_owner_users_access(self):
        self._test_add_with_file_perm_for_owner_users_access()

    def test_configure_add_chunk_size(self):
        self._test_configure_add_chunk_size()

    def test_multiple_backends_different_timeout_configs(self):
        """Test that multiple backends can have different timeout configs."""
        # Configure file1 with timeout enabled
        self.config(filesystem_store_timeout=30,
                    filesystem_store_thread_pool_size=5,
                    filesystem_store_threadpool_threshold=80,
                    group="file1")
        self.store.configure()

        # Verify file1 has correct timeout config
        self.assertEqual(self.store.timeout_executor.timeout, 30)
        self.assertEqual(self.store.timeout_executor.pool_size, 5)
        self.assertEqual(self.store.timeout_executor.threshold, 80)
        self.assertIsNotNone(self.store.timeout_executor.executor)

        # Configure file2 with different timeout config
        store2 = filesystem.Store(self.conf, backend='file2')
        test_dir2 = self.useFixture(fixtures.TempDir()).path
        self.config(filesystem_store_datadir=test_dir2,
                    filesystem_store_timeout=60,
                    filesystem_store_thread_pool_size=10,
                    filesystem_store_threadpool_threshold=90,
                    group="file2")
        store2.configure()

        # Verify file2 has different timeout config
        self.assertEqual(store2.timeout_executor.timeout, 60)
        self.assertEqual(store2.timeout_executor.pool_size, 10)
        self.assertEqual(store2.timeout_executor.threshold, 90)
        self.assertIsNotNone(store2.timeout_executor.executor)

        # Verify they are independent instances
        self.assertIsNot(self.store.timeout_executor,
                         store2.timeout_executor)

    def test_multiple_backends_one_with_timeout_one_without(self):
        """Test that one backend can have timeout enabled, another disabled."""
        # Configure file1 with timeout enabled
        self.config(filesystem_store_timeout=30, group="file1")
        self.store.configure()
        self.assertEqual(self.store.timeout_executor.timeout, 30)
        self.assertIsNotNone(self.store.timeout_executor.executor)

        # Configure file2 with timeout disabled
        store2 = filesystem.Store(self.conf, backend='file2')
        test_dir2 = self.useFixture(fixtures.TempDir()).path
        self.config(filesystem_store_datadir=test_dir2,
                    filesystem_store_timeout=0,
                    group="file2")
        store2.configure()

        # Verify file2 has timeout disabled
        self.assertEqual(store2.timeout_executor.timeout, 0)
        self.assertIsInstance(store2.timeout_executor.executor,
                              futurist.SynchronousExecutor)

    def test_backend_defaults_timeout_config(self):
        """Test that backend_defaults timeout config is used."""
        # Set timeout in backend_defaults BEFORE creating stores
        self.config(filesystem_store_timeout=45,
                    filesystem_store_thread_pool_size=8,
                    filesystem_store_threadpool_threshold=85,
                    group="backend_defaults")

        # Recreate store1 to pick up backend_defaults
        self.store = filesystem.Store(self.conf, backend='file1')
        self.config(filesystem_store_datadir=self.test_dir,
                    filesystem_store_chunk_size=10,
                    group="file1")
        self.store.configure()

        # Verify file1 uses backend_defaults values
        self.assertEqual(self.store.timeout_executor.timeout, 45)
        self.assertEqual(self.store.timeout_executor.pool_size, 8)
        self.assertEqual(self.store.timeout_executor.threshold, 85)
        self.assertIsNotNone(self.store.timeout_executor.executor)

        # Configure file2 with explicit timeout (should override
        # backend_defaults)
        store2 = filesystem.Store(self.conf, backend='file2')
        test_dir2 = self.useFixture(fixtures.TempDir()).path
        self.config(filesystem_store_datadir=test_dir2,
                    filesystem_store_timeout=20,
                    group="file2")
        store2.configure()

        # Verify file2 uses its own timeout, but pool_size and threshold
        # from backend_defaults
        self.assertEqual(store2.timeout_executor.timeout, 20)
        # From backend_defaults
        self.assertEqual(store2.timeout_executor.pool_size, 8)
        # From backend_defaults
        self.assertEqual(store2.timeout_executor.threshold, 85)

    def test_multiple_backends_timeout_operations_independent(self):
        """Test that timeout operations work independently for each backend."""
        # Configure both backends with timeout
        self.config(filesystem_store_timeout=1, group="file1")
        self.store.configure()

        store2 = filesystem.Store(self.conf, backend='file2')
        test_dir2 = self.useFixture(fixtures.TempDir()).path
        self.config(filesystem_store_datadir=test_dir2,
                    filesystem_store_timeout=1,
                    group="file2")
        store2.configure()

        # Create images in both backends
        image_id1 = str(uuid.uuid4())
        file_contents = b"test content 1"
        image_file1 = io.BytesIO(file_contents)
        uri1, size1, checksum1, multihash1 = self.store.add(
            image_id1, image_file1, len(file_contents))
        loc1 = location.get_location_from_uri_and_backend(
            uri1, 'file1', conf=self.conf)

        image_id2 = str(uuid.uuid4())
        file_contents2 = b"test content 2"
        image_file2 = io.BytesIO(file_contents2)
        uri2, size2, checksum2, multihash2 = store2.add(
            image_id2, image_file2, len(file_contents2))
        loc2 = location.get_location_from_uri_and_backend(
            uri2, 'file2', conf=self.conf)

        # Test timeout on file1 backend
        with mock.patch('os.unlink') as mock_unlink:
            def slow_unlink(path):
                time.sleep(2)
            mock_unlink.side_effect = slow_unlink
            self.assertRaises(exceptions.TimeoutError,
                              self.store.delete, loc1)

        # Test that file2 backend still works normally
        result_size = store2.get_size(loc2)
        self.assertEqual(result_size, len(file_contents2))
